# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.
import decimal
import errno
import os

from m3u8.mixins import BasePathMixin, GroupedBasePathMixin
from m3u8.parser import format_date_time, parse
from m3u8.protocol import (
    ext_oatcls_scte35,
    ext_x_asset,
    ext_x_key,
    ext_x_map,
    ext_x_session_key,
    ext_x_start,
)


class MalformedPlaylistError(Exception):
    pass


class M3U8(object):
    """
    Represents a single M3U8 playlist. Should be instantiated with
    the content as string.

    Parameters:

     `content`
       the m3u8 content as string

     `base_path`
       all urls (key and segments url) will be updated with this base_path,
       ex.:
           base_path = "http://videoserver.com/hls"

            /foo/bar/key.bin           -->  http://videoserver.com/hls/key.bin
            http://vid.com/segment1.ts -->  http://videoserver.com/hls/segment1.ts

       can be passed as parameter or setted as an attribute to ``M3U8`` object.
     `base_uri`
      uri the playlist comes from. it is propagated to SegmentList and Key
      ex.: http://example.com/path/to

    Attributes:

     `keys`
       Returns the list of `Key` objects used to encrypt the segments from m3u8.
       It covers the whole list of possible situations when encryption either is
       used or not.

       1. No encryption.
       `keys` list will only contain a `None` element.

       2. Encryption enabled for all segments.
       `keys` list will contain the key used for the segments.

       3. No encryption for first element(s), encryption is applied afterwards
       `keys` list will contain `None` and the key used for the rest of segments.

       4. Multiple keys used during the m3u8 manifest.
       `keys` list will contain the key used for each set of segments.

     `session_keys`
       Returns the list of `SessionKey` objects used to encrypt multiple segments from m3u8.

     `segments`
       a `SegmentList` object, represents the list of `Segment`s from this playlist

     `is_variant`
        Returns true if this M3U8 is a variant playlist, with links to
        other M3U8s with different bitrates.

        If true, `playlists` is a list of the playlists available,
        and `iframe_playlists` is a list of the i-frame playlists available.

     `is_endlist`
        Returns true if EXT-X-ENDLIST tag present in M3U8.
        http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.8

      `playlists`
        If this is a variant playlist (`is_variant` is True), returns a list of
        Playlist objects

      `iframe_playlists`
        If this is a variant playlist (`is_variant` is True), returns a list of
        IFramePlaylist objects

      `playlist_type`
        A lower-case string representing the type of the playlist, which can be
        one of VOD (video on demand) or EVENT.

      `media`
        If this is a variant playlist (`is_variant` is True), returns a list of
        Media objects

      `target_duration`
        Returns the EXT-X-TARGETDURATION as an integer
        http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.2

      `media_sequence`
        Returns the EXT-X-MEDIA-SEQUENCE as an integer
        http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.3

      `program_date_time`
        Returns the EXT-X-PROGRAM-DATE-TIME as a string
        http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.5

      `version`
        Return the EXT-X-VERSION as is

      `allow_cache`
        Return the EXT-X-ALLOW-CACHE as is

      `files`
        Returns an iterable with all files from playlist, in order. This includes
        segments and key uri, if present.

      `base_uri`
        It is a property (getter and setter) used by
        SegmentList and Key to have absolute URIs.

      `is_i_frames_only`
        Returns true if EXT-X-I-FRAMES-ONLY tag present in M3U8.
        http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.12

      `is_independent_segments`
        Returns true if EXT-X-INDEPENDENT-SEGMENTS tag present in M3U8.
        https://tools.ietf.org/html/draft-pantos-http-live-streaming-13#section-3.4.16

      `image_playlists`
        If this is a variant playlist (`is_variant` is True), returns a list of
        ImagePlaylist objects

      `is_images_only`
        Returns true if EXT-X-IMAGES-ONLY tag present in M3U8.
        https://github.com/image-media-playlist/spec/blob/master/image_media_playlist_v0_4.pdf
    """

    simple_attributes = (
        # obj attribute      # parser attribute
        ("is_variant", "is_variant"),
        ("is_endlist", "is_endlist"),
        ("is_i_frames_only", "is_i_frames_only"),
        ("target_duration", "targetduration"),
        ("media_sequence", "media_sequence"),
        ("program_date_time", "program_date_time"),
        ("is_independent_segments", "is_independent_segments"),
        ("version", "version"),
        ("allow_cache", "allow_cache"),
        ("playlist_type", "playlist_type"),
        ("discontinuity_sequence", "discontinuity_sequence"),
        ("is_images_only", "is_images_only")
    )

    def __init__(
        self,
        content=None,
        base_path=None,
        base_uri=None,
        strict=False,
        custom_tags_parser=None,
    ):
        if content is not None:
            self.data = parse(content, strict, custom_tags_parser)
        else:
            self.data = {}
        self._base_uri = base_uri
        if self._base_uri:
            if not self._base_uri.endswith("/"):
                self._base_uri += "/"

        self._initialize_attributes()
        self.base_path = base_path

    def _initialize_attributes(self):
        self.keys = [
            Key(base_uri=self.base_uri, **params) if params else None
            for params in self.data.get("keys", [])
        ]
        self.segment_map = [
            InitializationSection(base_uri=self.base_uri, **params) if params else None
            for params in self.data.get("segment_map", [])
        ]
        self.segments = SegmentList(
            [
                Segment(
                    base_uri=self.base_uri,
                    keyobject=find_key(segment.get("key", {}), self.keys),
                    **segment,
                )
                for segment in self.data.get("segments", [])
            ]
        )

        for attr, param in self.simple_attributes:
            setattr(self, attr, self.data.get(param))

        for i, segment in enumerate(self.segments, self.media_sequence or 0):
            segment.media_sequence = i

        self.files = []
        for key in self.keys:
            # Avoid None key, it could be the first one, don't repeat them
            if key and key.uri not in self.files:
                self.files.append(key.uri)
        self.files.extend(self.segments.uri)

        self.media = MediaList(
            [
                Media(base_uri=self.base_uri, **media)
                for media in self.data.get("media", [])
            ]
        )

        self.playlists = PlaylistList(
            [
                Playlist(base_uri=self.base_uri, media=self.media, **playlist)
                for playlist in self.data.get("playlists", [])
            ]
        )

        self.iframe_playlists = PlaylistList()
        for ifr_pl in self.data.get("iframe_playlists", []):
            self.iframe_playlists.append(
                IFramePlaylist(
                    base_uri=self.base_uri,
                    uri=ifr_pl["uri"],
                    iframe_stream_info=ifr_pl["iframe_stream_info"],
                )
            )

        self.image_playlists = PlaylistList()
        for img_pl in self.data.get('image_playlists', []):
            self.image_playlists.append(
                ImagePlaylist(
                    base_uri=self.base_uri,
                    uri=img_pl["uri"],
                    image_stream_info=img_pl["image_stream_info"]
                )
            )

        start = self.data.get("start", None)
        self.start = start and Start(**start)

        server_control = self.data.get("server_control", None)
        self.server_control = server_control and ServerControl(**server_control)

        part_inf = self.data.get("part_inf", None)
        self.part_inf = part_inf and PartInformation(**part_inf)

        skip = self.data.get("skip", None)
        self.skip = skip and Skip(**skip)

        self.rendition_reports = RenditionReportList(
            [
                RenditionReport(base_uri=self.base_uri, **rendition_report)
                for rendition_report in self.data.get("rendition_reports", [])
            ]
        )

        self.session_data = SessionDataList(
            [
                SessionData(**session_data)
                for session_data in self.data.get("session_data", [])
                if "data_id" in session_data
            ]
        )

        self.session_keys = [
            SessionKey(base_uri=self.base_uri, **params) if params else None
            for params in self.data.get("session_keys", [])
        ]

        preload_hint = self.data.get("preload_hint", None)
        self.preload_hint = preload_hint and PreloadHint(
            base_uri=self.base_uri, **preload_hint
        )

        content_steering = self.data.get("content_steering", None)
        self.content_steering = content_steering and ContentSteering(
            base_uri=self.base_uri, **content_steering
        )

    def __unicode__(self):
        return self.dumps()

    @property
    def base_uri(self):
        return self._base_uri

    @base_uri.setter
    def base_uri(self, new_base_uri):
        self._base_uri = new_base_uri
        self.media.base_uri = new_base_uri
        self.playlists.base_uri = new_base_uri
        self.iframe_playlists.base_uri = new_base_uri
        self.segments.base_uri = new_base_uri
        self.rendition_reports.base_uri = new_base_uri
        self.image_playlists.base_uri = new_base_uri
        for key in self.keys:
            if key:
                key.base_uri = new_base_uri
        for key in self.session_keys:
            if key:
                key.base_uri = new_base_uri
        if self.preload_hint:
            self.preload_hint.base_uri = new_base_uri
        if self.content_steering:
            self.content_steering.base_uri = new_base_uri

    @property
    def base_path(self):
        return self._base_path

    @base_path.setter
    def base_path(self, newbase_path):
        self._base_path = newbase_path
        self._update_base_path()

    def _update_base_path(self):
        if self._base_path is None:
            return
        for key in self.keys:
            if key:
                key.base_path = self._base_path
        for key in self.session_keys:
            if key:
                key.base_path = self._base_path
        self.media.base_path = self._base_path
        self.segments.base_path = self._base_path
        self.playlists.base_path = self._base_path
        self.iframe_playlists.base_path = self._base_path
        self.image_playlists.base_path = self._base_path
        self.rendition_reports.base_path = self._base_path
        if self.preload_hint:
            self.preload_hint.base_path = self._base_path
        if self.content_steering:
            self.content_steering.base_path = self._base_path

    def add_playlist(self, playlist):
        self.is_variant = True
        self.playlists.append(playlist)

    def add_iframe_playlist(self, iframe_playlist):
        if iframe_playlist is not None:
            self.is_variant = True
            self.iframe_playlists.append(iframe_playlist)

    def add_image_playlist(self, image_playlist):
        if image_playlist is not None:
            self.is_variant = True
            self.image_playlists.append(image_playlist)

    def add_media(self, media):
        self.media.append(media)

    def add_segment(self, segment):
        self.segments.append(segment)

    def add_rendition_report(self, report):
        self.rendition_reports.append(report)

    def dumps(self):
        """
        Returns the current m3u8 as a string.
        You could also use unicode(<this obj>) or str(<this obj>)
        """
        output = ["#EXTM3U"]
        if self.content_steering:
            output.append(str(self.content_steering))
        if self.media_sequence:
            output.append("#EXT-X-MEDIA-SEQUENCE:" + str(self.media_sequence))
        if self.discontinuity_sequence:
            output.append(
                "#EXT-X-DISCONTINUITY-SEQUENCE:{}".format(self.discontinuity_sequence)
            )
        if self.allow_cache:
            output.append("#EXT-X-ALLOW-CACHE:" + self.allow_cache.upper())
        if self.version:
            output.append("#EXT-X-VERSION:" + str(self.version))
        if self.is_independent_segments:
            output.append("#EXT-X-INDEPENDENT-SEGMENTS")
        if self.target_duration:
            output.append(
                "#EXT-X-TARGETDURATION:" + number_to_string(self.target_duration)
            )
        if not (self.playlist_type is None or self.playlist_type == ""):
            output.append("#EXT-X-PLAYLIST-TYPE:%s" % str(self.playlist_type).upper())
        if self.start:
            output.append(str(self.start))
        if self.is_i_frames_only:
            output.append("#EXT-X-I-FRAMES-ONLY")
        if self.is_images_only:
            output.append("#EXT-X-IMAGES-ONLY")
        if self.server_control:
            output.append(str(self.server_control))
        if self.is_variant:
            if self.media:
                output.append(str(self.media))
            output.append(str(self.playlists))
            if self.iframe_playlists:
                output.append(str(self.iframe_playlists))
            if self.image_playlists:
                output.append(str(self.image_playlists))
        if self.part_inf:
            output.append(str(self.part_inf))
        if self.skip:
            output.append(str(self.skip))
        if self.session_data:
            output.append(str(self.session_data))

        for key in self.session_keys:
            output.append(str(key))

        output.append(str(self.segments))

        if self.preload_hint:
            output.append(str(self.preload_hint))

        if self.rendition_reports:
            output.append(str(self.rendition_reports))

        if self.is_endlist:
            output.append("#EXT-X-ENDLIST")

        # ensure that the last line is terminated correctly
        if output[-1] and not output[-1].endswith("\n"):
            output.append("")

        return "\n".join(output)

    def dump(self, filename):
        """
        Saves the current m3u8 to ``filename``
        """
        self._create_sub_directories(filename)

        with open(filename, "w") as fileobj:
            fileobj.write(self.dumps())

    def _create_sub_directories(self, filename):
        basename = os.path.dirname(filename)
        if basename:
            os.makedirs(basename, exist_ok=True)


class Segment(BasePathMixin):
    """
    A video segment from a M3U8 playlist

    `uri`
      a string with the segment uri

    `title`
      title attribute from EXTINF parameter

    `program_date_time`
      Returns the EXT-X-PROGRAM-DATE-TIME as a datetime. This field is only set
      if EXT-X-PROGRAM-DATE-TIME exists for this segment
      http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.5

    `current_program_date_time`
      Returns a datetime of this segment, either the value of `program_date_time`
      when EXT-X-PROGRAM-DATE-TIME is set or a calculated value based on previous
      segments' EXT-X-PROGRAM-DATE-TIME and EXTINF values

    `discontinuity`
      Returns a boolean indicating if a EXT-X-DISCONTINUITY tag exists
      http://tools.ietf.org/html/draft-pantos-http-live-streaming-13#section-3.4.11

    `cue_out_start`
      Returns a boolean indicating if a EXT-X-CUE-OUT tag exists

    `cue_out`
      Returns a boolean indicating if a EXT-X-CUE-OUT-CONT tag exists
      Note: for backwards compatibility, this will be True when cue_out_start
            is True, even though this tag did not exist in the input, and
            EXT-X-CUE-OUT-CONT will not exist in the output

    `cue_in`
      Returns a boolean indicating if a EXT-X-CUE-IN tag exists

    `scte35`
      Base64 encoded SCTE35 metadata if available

    `scte35_duration`
      Planned SCTE35 duration

    `duration`
      duration attribute from EXTINF parameter

    `base_uri`
      uri the key comes from in URI hierarchy. ex.: http://example.com/path/to

    `bitrate`
      bitrate attribute from EXT-X-BITRATE parameter

    `byterange`
      byterange attribute from EXT-X-BYTERANGE parameter

    `key`
      Key used to encrypt the segment (EXT-X-KEY)

    `parts`
      partial segments that make up this segment

    `dateranges`
      any dateranges that should  precede the segment

    `gap_tag`
      GAP tag indicates that a Media Segment is missing

    `custom_parser_values`
        Additional values which custom_tags_parser might store per segment
    """

    def __init__(
        self,
        uri=None,
        base_uri=None,
        program_date_time=None,
        current_program_date_time=None,
        duration=None,
        title=None,
        bitrate=None,
        byterange=None,
        cue_out=False,
        cue_out_start=False,
        cue_in=False,
        discontinuity=False,
        key=None,
        scte35=None,
        oatcls_scte35=None,
        scte35_duration=None,
        scte35_elapsedtime=None,
        asset_metadata=None,
        keyobject=None,
        parts=None,
        init_section=None,
        dateranges=None,
        gap_tag=None,
        media_sequence=None,
        custom_parser_values=None,
    ):
        self.media_sequence = media_sequence
        self.uri = uri
        self.duration = duration
        self.title = title
        self._base_uri = base_uri
        self.bitrate = bitrate
        self.byterange = byterange
        self.program_date_time = program_date_time
        self.current_program_date_time = current_program_date_time
        self.discontinuity = discontinuity
        self.cue_out_start = cue_out_start
        self.cue_out = cue_out
        self.cue_in = cue_in
        self.scte35 = scte35
        self.oatcls_scte35 = oatcls_scte35
        self.scte35_duration = scte35_duration
        self.scte35_elapsedtime = scte35_elapsedtime
        self.asset_metadata = asset_metadata
        self.key = keyobject
        self.parts = PartialSegmentList(
            [PartialSegment(base_uri=self._base_uri, **partial) for partial in parts]
            if parts
            else []
        )
        if init_section is not None:
            self.init_section = InitializationSection(self._base_uri, **init_section)
        else:
            self.init_section = None
        self.dateranges = DateRangeList(
            [DateRange(**daterange) for daterange in dateranges] if dateranges else []
        )
        self.gap_tag = gap_tag
        self.custom_parser_values = custom_parser_values or {}

    def add_part(self, part):
        self.parts.append(part)

    def dumps(self, last_segment, timespec="milliseconds"):
        output = []

        if last_segment and self.key != last_segment.key:
            output.append(str(self.key))
            output.append("\n")
        else:
            # The key must be checked anyway now for the first segment
            if self.key and last_segment is None:
                output.append(str(self.key))
                output.append("\n")

        if last_segment and self.init_section != last_segment.init_section:
            if not self.init_section:
                raise MalformedPlaylistError(
                    "init section can't be None if previous is not None"
                )
            output.append(str(self.init_section))
            output.append("\n")
        else:
            if self.init_section and last_segment is None:
                output.append(str(self.init_section))
                output.append("\n")

        if self.discontinuity:
            output.append("#EXT-X-DISCONTINUITY\n")
        if self.program_date_time:
            output.append(
                "#EXT-X-PROGRAM-DATE-TIME:%s\n"
                % format_date_time(self.program_date_time, timespec=timespec)
            )

        if len(self.dateranges):
            output.append(str(self.dateranges))
            output.append("\n")

        if self.cue_out_start:
            if self.oatcls_scte35:
                output.append(f"{ext_oatcls_scte35}:{self.oatcls_scte35}\n")

            if self.asset_metadata:
                asset_suffix = []
                for metadata_key, metadata_value in self.asset_metadata.items():
                    asset_suffix.append(f"{metadata_key.upper()}={metadata_value}")
                output.append(f"{ext_x_asset}:{','.join(asset_suffix)}\n")

            output.append(
                "#EXT-X-CUE-OUT{}\n".format(
                    (":" + self.scte35_duration) if self.scte35_duration else ""
                )
            )
        elif self.cue_out:
            cue_out_cont_suffix = []
            if self.scte35_elapsedtime:
                cue_out_cont_suffix.append(f"ElapsedTime={self.scte35_elapsedtime}")
            if self.scte35_duration:
                cue_out_cont_suffix.append(f"Duration={self.scte35_duration}")
            if self.scte35:
                cue_out_cont_suffix.append(f"SCTE35={self.scte35}")

            if cue_out_cont_suffix:
                cue_out_cont_suffix = ":" + ",".join(cue_out_cont_suffix)
            else:
                cue_out_cont_suffix = ""
            output.append(f"#EXT-X-CUE-OUT-CONT{cue_out_cont_suffix}\n")
        if self.cue_in:
            output.append("#EXT-X-CUE-IN\n")

        if self.parts:
            output.append(str(self.parts))
            output.append("\n")

        if self.uri:
            if self.duration is not None:
                output.append("#EXTINF:%s," % number_to_string(self.duration))
                if self.title:
                    output.append(self.title)
                output.append("\n")

            if self.byterange:
                output.append("#EXT-X-BYTERANGE:%s\n" % self.byterange)

            if self.bitrate:
                output.append("#EXT-X-BITRATE:%s\n" % self.bitrate)

            if self.gap_tag:
                output.append("#EXT-X-GAP\n")

            output.append(self.uri)

        return "".join(output)

    def __str__(self):
        return self.dumps(None)

    @property
    def base_path(self):
        return super(Segment, self).base_path

    @base_path.setter
    def base_path(self, newbase_path):
        super(Segment, self.__class__).base_path.fset(self, newbase_path)
        self.parts.base_path = newbase_path
        if self.init_section is not None:
            self.init_section.base_path = newbase_path

    @property
    def base_uri(self):
        return self._base_uri

    @base_uri.setter
    def base_uri(self, newbase_uri):
        self._base_uri = newbase_uri
        self.parts.base_uri = newbase_uri
        if self.init_section is not None:
            self.init_section.base_uri = newbase_uri


class SegmentList(list, GroupedBasePathMixin):
    def __str__(self):
        output = []
        last_segment = None
        for segment in self:
            output.append(segment.dumps(last_segment))
            last_segment = segment
        return "\n".join(output)

    @property
    def uri(self):
        return [seg.uri for seg in self]

    def by_key(self, key):
        return [segment for segment in self if segment.key == key]


class PartialSegment(BasePathMixin):
    """
    A partial segment from a M3U8 playlist

    `uri`
      a string with the segment uri

    `program_date_time`
      Returns the EXT-X-PROGRAM-DATE-TIME as a datetime. This field is only set
      if EXT-X-PROGRAM-DATE-TIME exists for this segment
      http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.5

    `current_program_date_time`
      Returns a datetime of this segment, either the value of `program_date_time`
      when EXT-X-PROGRAM-DATE-TIME is set or a calculated value based on previous
      segments' EXT-X-PROGRAM-DATE-TIME and EXTINF values

    `duration`
      duration attribute from EXTINF parameter

    `byterange`
      byterange attribute from EXT-X-BYTERANGE parameter

    `independent`
      the Partial Segment contains an independent frame

    `gap`
      GAP attribute indicates the Partial Segment is not available

    `dateranges`
      any dateranges that should precede the partial segment

    `gap_tag`
      GAP tag indicates one or more of the parent Media Segment's Partial
      Segments have a GAP=YES attribute. This tag should appear immediately
      after the first EXT-X-PART tag in the Parent Segment with a GAP=YES
      attribute.
    """

    def __init__(
        self,
        base_uri,
        uri,
        duration,
        program_date_time=None,
        current_program_date_time=None,
        byterange=None,
        independent=None,
        gap=None,
        dateranges=None,
        gap_tag=None,
    ):
        self.base_uri = base_uri
        self.uri = uri
        self.duration = duration
        self.program_date_time = program_date_time
        self.current_program_date_time = current_program_date_time
        self.byterange = byterange
        self.independent = independent
        self.gap = gap
        self.dateranges = DateRangeList(
            [DateRange(**daterange) for daterange in dateranges] if dateranges else []
        )
        self.gap_tag = gap_tag

    def dumps(self, last_segment):
        output = []

        if len(self.dateranges):
            output.append(str(self.dateranges))
            output.append("\n")

        if self.gap_tag:
            output.append("#EXT-X-GAP\n")

        output.append(
            '#EXT-X-PART:DURATION=%s,URI="%s"'
            % (number_to_string(self.duration), self.uri)
        )

        if self.independent:
            output.append(",INDEPENDENT=%s" % self.independent)

        if self.byterange:
            output.append(",BYTERANGE=%s" % self.byterange)

        if self.gap:
            output.append(",GAP=%s" % self.gap)

        return "".join(output)

    def __str__(self):
        return self.dumps(None)


class PartialSegmentList(list, GroupedBasePathMixin):
    def __str__(self):
        output = [str(part) for part in self]
        return "\n".join(output)


class Key(BasePathMixin):
    """
    Key used to encrypt the segments in a m3u8 playlist (EXT-X-KEY)

    `method`
      is a string. ex.: "AES-128"

    `uri`
      is a string. ex:: "https://priv.example.com/key.php?r=52"

    `base_uri`
      uri the key comes from in URI hierarchy. ex.: http://example.com/path/to

    `iv`
      initialization vector. a string representing a hexadecimal number. ex.: 0X12A

    """

    tag = ext_x_key

    def __init__(
        self,
        method,
        base_uri,
        uri=None,
        iv=None,
        keyformat=None,
        keyformatversions=None,
        **kwargs,
    ):
        self.method = method
        self.uri = uri
        self.iv = iv
        self.keyformat = keyformat
        self.keyformatversions = keyformatversions
        self.base_uri = base_uri
        self._extra_params = kwargs

    def __str__(self):
        output = [
            "METHOD=%s" % self.method,
        ]
        if self.uri:
            output.append('URI="%s"' % self.uri)
        if self.iv:
            output.append("IV=%s" % self.iv)
        if self.keyformat:
            output.append('KEYFORMAT="%s"' % self.keyformat)
        if self.keyformatversions:
            output.append('KEYFORMATVERSIONS="%s"' % self.keyformatversions)

        return self.tag + ":" + ",".join(output)

    def __eq__(self, other):
        if not other:
            return False
        return (
            self.method == other.method
            and self.uri == other.uri
            and self.iv == other.iv
            and self.base_uri == other.base_uri
            and self.keyformat == other.keyformat
            and self.keyformatversions == other.keyformatversions
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class InitializationSection(BasePathMixin):
    """
    Used to obtain Media Initialization Section required to
    parse the applicable Media Segments (EXT-X-MAP)

    `uri`
      is a string. ex:: "https://priv.example.com/key.php?r=52"

    `byterange`
      value of BYTERANGE attribute

    `base_uri`
      uri the segment comes from in URI hierarchy. ex.: http://example.com/path/to
    """

    tag = ext_x_map

    def __init__(self, base_uri, uri, byterange=None):
        self.base_uri = base_uri
        self.uri = uri
        self.byterange = byterange

    def __str__(self):
        output = []
        if self.uri:
            output.append("URI=" + quoted(self.uri))
        if self.byterange:
            output.append("BYTERANGE=" + quoted(self.byterange))
        return "{tag}:{attributes}".format(tag=self.tag, attributes=",".join(output))

    def __eq__(self, other):
        if not other:
            return False
        return (
            self.uri == other.uri
            and self.byterange == other.byterange
            and self.base_uri == other.base_uri
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class SessionKey(Key):
    tag = ext_x_session_key


class Playlist(BasePathMixin):
    """
    Playlist object representing a link to a variant M3U8 with a specific bitrate.

    Attributes:

    `stream_info` is a named tuple containing the attributes: `program_id`,
    `bandwidth`, `average_bandwidth`, `resolution`, `codecs` and `resolution`
    which is a a tuple (w, h) of integers

    `media` is a list of related Media entries.

    More info: http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.10
    """

    def __init__(self, uri, stream_info, media, base_uri):
        self.uri = uri
        self.base_uri = base_uri

        resolution = stream_info.get("resolution")
        if resolution is not None:
            resolution = resolution.strip('"')
            values = resolution.split("x")
            resolution_pair = (int(values[0]), int(values[1]))
        else:
            resolution_pair = None

        self.stream_info = StreamInfo(
            bandwidth=stream_info["bandwidth"],
            video=stream_info.get("video"),
            audio=stream_info.get("audio"),
            subtitles=stream_info.get("subtitles"),
            closed_captions=stream_info.get("closed_captions"),
            average_bandwidth=stream_info.get("average_bandwidth"),
            program_id=stream_info.get("program_id"),
            resolution=resolution_pair,
            codecs=stream_info.get("codecs"),
            frame_rate=stream_info.get("frame_rate"),
            video_range=stream_info.get("video_range"),
            hdcp_level=stream_info.get("hdcp_level"),
            pathway_id=stream_info.get("pathway_id"),
            stable_variant_id=stream_info.get("stable_variant_id"),
        )
        self.media = []
        for media_type in ("audio", "video", "subtitles"):
            group_id = stream_info.get(media_type)
            if not group_id:
                continue

            self.media += filter(lambda m: m.group_id == group_id, media)

    def __str__(self):
        media_types = []
        stream_inf = [str(self.stream_info)]
        for media in self.media:
            if media.type in media_types:
                continue
            else:
                media_types += [media.type]
                media_type = media.type.upper()
                stream_inf.append('%s="%s"' % (media_type, media.group_id))

        return "#EXT-X-STREAM-INF:" + ",".join(stream_inf) + "\n" + self.uri


class IFramePlaylist(BasePathMixin):
    """
    IFramePlaylist object representing a link to a
    variant M3U8 i-frame playlist with a specific bitrate.

    Attributes:

    `iframe_stream_info` is a named tuple containing the attributes:
     `program_id`, `bandwidth`, `average_bandwidth`, `codecs`, `video_range`,
     `hdcp_level` and `resolution` which is a tuple (w, h) of integers

    More info: http://tools.ietf.org/html/draft-pantos-http-live-streaming-07#section-3.3.13
    """

    def __init__(self, base_uri, uri, iframe_stream_info):
        self.uri = uri
        self.base_uri = base_uri

        resolution = iframe_stream_info.get("resolution")
        if resolution is not None:
            values = resolution.split("x")
            resolution_pair = (int(values[0]), int(values[1]))
        else:
            resolution_pair = None

        self.iframe_stream_info = StreamInfo(
            bandwidth=iframe_stream_info.get("bandwidth"),
            average_bandwidth=iframe_stream_info.get("average_bandwidth"),
            video=iframe_stream_info.get("video"),
            # Audio, subtitles, and closed captions should not exist in
            # EXT-X-I-FRAME-STREAM-INF, so just hardcode them to None.
            audio=None,
            subtitles=None,
            closed_captions=None,
            program_id=iframe_stream_info.get("program_id"),
            resolution=resolution_pair,
            codecs=iframe_stream_info.get("codecs"),
            video_range=iframe_stream_info.get("video_range"),
            hdcp_level=iframe_stream_info.get("hdcp_level"),
            frame_rate=None,
            pathway_id=iframe_stream_info.get("pathway_id"),
            stable_variant_id=iframe_stream_info.get("stable_variant_id"),
        )

    def __str__(self):
        iframe_stream_inf = []
        if self.iframe_stream_info.program_id:
            iframe_stream_inf.append(
                "PROGRAM-ID=%d" % self.iframe_stream_info.program_id
            )
        if self.iframe_stream_info.bandwidth:
            iframe_stream_inf.append("BANDWIDTH=%d" % self.iframe_stream_info.bandwidth)
        if self.iframe_stream_info.average_bandwidth:
            iframe_stream_inf.append(
                "AVERAGE-BANDWIDTH=%d" % self.iframe_stream_info.average_bandwidth
            )
        if self.iframe_stream_info.resolution:
            res = (
                str(self.iframe_stream_info.resolution[0])
                + "x"
                + str(self.iframe_stream_info.resolution[1])
            )
            iframe_stream_inf.append("RESOLUTION=" + res)
        if self.iframe_stream_info.codecs:
            iframe_stream_inf.append("CODECS=" + quoted(self.iframe_stream_info.codecs))
        if self.iframe_stream_info.video_range:
            iframe_stream_inf.append(
                "VIDEO-RANGE=%s" % self.iframe_stream_info.video_range
            )
        if self.iframe_stream_info.hdcp_level:
            iframe_stream_inf.append(
                "HDCP-LEVEL=%s" % self.iframe_stream_info.hdcp_level
            )
        if self.uri:
            iframe_stream_inf.append("URI=" + quoted(self.uri))
        if self.iframe_stream_info.pathway_id:
            iframe_stream_inf.append(
                "PATHWAY-ID=" + quoted(self.iframe_stream_info.pathway_id)
            )
        if self.iframe_stream_info.stable_variant_id:
            iframe_stream_inf.append(
                "STABLE-VARIANT-ID=" + quoted(self.iframe_stream_info.stable_variant_id)
            )

        return "#EXT-X-I-FRAME-STREAM-INF:" + ",".join(iframe_stream_inf)


class StreamInfo(object):
    bandwidth = None
    closed_captions = None
    average_bandwidth = None
    program_id = None
    resolution = None
    codecs = None
    audio = None
    video = None
    subtitles = None
    frame_rate = None
    video_range = None
    hdcp_level = None
    pathway_id = None
    stable_variant_id = None

    def __init__(self, **kwargs):
        self.bandwidth = kwargs.get("bandwidth")
        self.closed_captions = kwargs.get("closed_captions")
        self.average_bandwidth = kwargs.get("average_bandwidth")
        self.program_id = kwargs.get("program_id")
        self.resolution = kwargs.get("resolution")
        self.codecs = kwargs.get("codecs")
        self.audio = kwargs.get("audio")
        self.video = kwargs.get("video")
        self.subtitles = kwargs.get("subtitles")
        self.frame_rate = kwargs.get("frame_rate")
        self.video_range = kwargs.get("video_range")
        self.hdcp_level = kwargs.get("hdcp_level")
        self.pathway_id = kwargs.get("pathway_id")
        self.stable_variant_id = kwargs.get("stable_variant_id")

    def __str__(self):
        stream_inf = []
        if self.program_id is not None:
            stream_inf.append("PROGRAM-ID=%d" % self.program_id)
        if self.closed_captions is not None:
            stream_inf.append("CLOSED-CAPTIONS=%s" % self.closed_captions)
        if self.bandwidth is not None:
            stream_inf.append("BANDWIDTH=%d" % self.bandwidth)
        if self.average_bandwidth is not None:
            stream_inf.append("AVERAGE-BANDWIDTH=%d" % self.average_bandwidth)
        if self.resolution is not None:
            res = str(self.resolution[0]) + "x" + str(self.resolution[1])
            stream_inf.append("RESOLUTION=" + res)
        if self.frame_rate is not None:
            stream_inf.append(
                "FRAME-RATE=%g"
                % decimal.Decimal(self.frame_rate).quantize(decimal.Decimal("1.000"))
            )
        if self.codecs is not None:
            stream_inf.append("CODECS=" + quoted(self.codecs))
        if self.video_range is not None:
            stream_inf.append("VIDEO-RANGE=%s" % self.video_range)
        if self.hdcp_level is not None:
            stream_inf.append("HDCP-LEVEL=%s" % self.hdcp_level)
        if self.pathway_id is not None:
            stream_inf.append("PATHWAY-ID=" + quoted(self.pathway_id))
        if self.stable_variant_id is not None:
            stream_inf.append("STABLE-VARIANT-ID=" + quoted(self.stable_variant_id))
        return ",".join(stream_inf)


class Media(BasePathMixin):
    """
    A media object from a M3U8 playlist
    https://tools.ietf.org/html/draft-pantos-http-live-streaming-16#section-4.3.4.1

    `uri`
      a string with the media uri

    `type`
    `group_id`
    `language`
    `assoc-language`
    `name`
    `default`
    `autoselect`
    `forced`
    `instream_id`
    `characteristics`
    `channels`
    `stable_rendition_id`
      attributes in the EXT-MEDIA tag

    `base_uri`
      uri the media comes from in URI hierarchy. ex.: http://example.com/path/to
    """

    def __init__(
        self,
        uri=None,
        type=None,
        group_id=None,
        language=None,
        name=None,
        default=None,
        autoselect=None,
        forced=None,
        characteristics=None,
        channels=None,
        stable_rendition_id=None,
        assoc_language=None,
        instream_id=None,
        base_uri=None,
        **extras,
    ):
        self.base_uri = base_uri
        self.uri = uri
        self.type = type
        self.group_id = group_id
        self.language = language
        self.name = name
        self.default = default
        self.autoselect = autoselect
        self.forced = forced
        self.assoc_language = assoc_language
        self.instream_id = instream_id
        self.characteristics = characteristics
        self.channels = channels
        self.stable_rendition_id = stable_rendition_id
        self.extras = extras

    def dumps(self):
        media_out = []

        if self.uri:
            media_out.append("URI=" + quoted(self.uri))
        if self.type:
            media_out.append("TYPE=" + self.type)
        if self.group_id:
            media_out.append("GROUP-ID=" + quoted(self.group_id))
        if self.language:
            media_out.append("LANGUAGE=" + quoted(self.language))
        if self.assoc_language:
            media_out.append("ASSOC-LANGUAGE=" + quoted(self.assoc_language))
        if self.name:
            media_out.append("NAME=" + quoted(self.name))
        if self.default:
            media_out.append("DEFAULT=" + self.default)
        if self.autoselect:
            media_out.append("AUTOSELECT=" + self.autoselect)
        if self.forced:
            media_out.append("FORCED=" + self.forced)
        if self.instream_id:
            media_out.append("INSTREAM-ID=" + quoted(self.instream_id))
        if self.characteristics:
            media_out.append("CHARACTERISTICS=" + quoted(self.characteristics))
        if self.channels:
            media_out.append("CHANNELS=" + quoted(self.channels))
        if self.stable_rendition_id:
            media_out.append("STABLE-RENDITION-ID=" + quoted(self.stable_rendition_id))

        return "#EXT-X-MEDIA:" + ",".join(media_out)

    def __str__(self):
        return self.dumps()


class TagList(list):
    def __str__(self):
        output = [str(tag) for tag in self]
        return "\n".join(output)


class MediaList(TagList, GroupedBasePathMixin):
    @property
    def uri(self):
        return [media.uri for media in self]


class PlaylistList(TagList, GroupedBasePathMixin):
    pass


class SessionDataList(TagList):
    pass


class Start(object):
    def __init__(self, time_offset, precise=None):
        self.time_offset = float(time_offset)
        self.precise = precise

    def __str__(self):
        output = ["TIME-OFFSET=" + str(self.time_offset)]
        if self.precise and self.precise in ["YES", "NO"]:
            output.append("PRECISE=" + str(self.precise))

        return ext_x_start + ":" + ",".join(output)


class RenditionReport(BasePathMixin):
    def __init__(self, base_uri, uri, last_msn, last_part=None):
        self.base_uri = base_uri
        self.uri = uri
        self.last_msn = last_msn
        self.last_part = last_part

    def dumps(self):
        report = []
        report.append("URI=" + quoted(self.uri))
        report.append("LAST-MSN=" + str(self.last_msn))
        if self.last_part is not None:
            report.append("LAST-PART=" + str(self.last_part))

        return "#EXT-X-RENDITION-REPORT:" + ",".join(report)

    def __str__(self):
        return self.dumps()


class RenditionReportList(list, GroupedBasePathMixin):
    def __str__(self):
        output = [str(report) for report in self]
        return "\n".join(output)


class ServerControl(object):
    def __init__(
        self,
        can_skip_until=None,
        can_block_reload=None,
        hold_back=None,
        part_hold_back=None,
        can_skip_dateranges=None,
    ):
        self.can_skip_until = can_skip_until
        self.can_block_reload = can_block_reload
        self.hold_back = hold_back
        self.part_hold_back = part_hold_back
        self.can_skip_dateranges = can_skip_dateranges

    def __getitem__(self, item):
        return getattr(self, item)

    def dumps(self):
        ctrl = []
        if self.can_block_reload:
            ctrl.append("CAN-BLOCK-RELOAD=%s" % self.can_block_reload)

        for attr in ["hold_back", "part_hold_back"]:
            if self[attr]:
                ctrl.append(
                    "%s=%s"
                    % (denormalize_attribute(attr), number_to_string(self[attr]))
                )

        if self.can_skip_until:
            ctrl.append("CAN-SKIP-UNTIL=%s" % number_to_string(self.can_skip_until))
            if self.can_skip_dateranges:
                ctrl.append("CAN-SKIP-DATERANGES=%s" % self.can_skip_dateranges)

        return "#EXT-X-SERVER-CONTROL:" + ",".join(ctrl)

    def __str__(self):
        return self.dumps()


class Skip(object):
    def __init__(self, skipped_segments, recently_removed_dateranges=None):
        self.skipped_segments = skipped_segments
        self.recently_removed_dateranges = recently_removed_dateranges

    def dumps(self):
        skip = []
        skip.append("SKIPPED-SEGMENTS=%s" % self.skipped_segments)
        if self.recently_removed_dateranges is not None:
            skip.append(
                "RECENTLY-REMOVED-DATERANGES=%s"
                % quoted(self.recently_removed_dateranges)
            )

        return "#EXT-X-SKIP:" + ",".join(skip)

    def __str__(self):
        return self.dumps()


class PartInformation(object):
    def __init__(self, part_target=None):
        self.part_target = part_target

    def dumps(self):
        return "#EXT-X-PART-INF:PART-TARGET=%s" % number_to_string(self.part_target)

    def __str__(self):
        return self.dumps()


class PreloadHint(BasePathMixin):
    def __init__(
        self, type, base_uri, uri, byterange_start=None, byterange_length=None
    ):
        self.hint_type = type
        self.base_uri = base_uri
        self.uri = uri
        self.byterange_start = byterange_start
        self.byterange_length = byterange_length

    def __getitem__(self, item):
        return getattr(self, item)

    def dumps(self):
        hint = []
        hint.append("TYPE=" + self.hint_type)
        hint.append("URI=" + quoted(self.uri))

        for attr in ["byterange_start", "byterange_length"]:
            if self[attr] is not None:
                hint.append("%s=%s" % (denormalize_attribute(attr), self[attr]))

        return "#EXT-X-PRELOAD-HINT:" + ",".join(hint)

    def __str__(self):
        return self.dumps()


class SessionData(object):
    def __init__(self, data_id, value=None, uri=None, language=None):
        self.data_id = data_id
        self.value = value
        self.uri = uri
        self.language = language

    def dumps(self):
        session_data_out = ["DATA-ID=" + quoted(self.data_id)]

        if self.value:
            session_data_out.append("VALUE=" + quoted(self.value))
        elif self.uri:
            session_data_out.append("URI=" + quoted(self.uri))
        if self.language:
            session_data_out.append("LANGUAGE=" + quoted(self.language))

        return "#EXT-X-SESSION-DATA:" + ",".join(session_data_out)

    def __str__(self):
        return self.dumps()


class DateRangeList(TagList):
    pass


class DateRange(object):
    def __init__(self, **kwargs):
        self.id = kwargs["id"]
        self.start_date = kwargs.get("start_date")
        self.class_ = kwargs.get("class")
        self.end_date = kwargs.get("end_date")
        self.duration = kwargs.get("duration")
        self.planned_duration = kwargs.get("planned_duration")
        self.scte35_cmd = kwargs.get("scte35_cmd")
        self.scte35_out = kwargs.get("scte35_out")
        self.scte35_in = kwargs.get("scte35_in")
        self.end_on_next = kwargs.get("end_on_next")
        self.x_client_attrs = [
            (attr, kwargs.get(attr)) for attr in kwargs if attr.startswith("x_")
        ]

    def dumps(self):
        daterange = []
        daterange.append("ID=" + quoted(self.id))

        # whilst START-DATE is technically REQUIRED by the spec, this is
        # contradicted by an example in the same document (see
        # https://tools.ietf.org/html/rfc8216#section-8.10), and also by
        # real-world implementations, so we make it optional here
        if self.start_date:
            daterange.append("START-DATE=" + quoted(self.start_date))
        if self.class_:
            daterange.append("CLASS=" + quoted(self.class_))
        if self.end_date:
            daterange.append("END-DATE=" + quoted(self.end_date))
        if self.duration:
            daterange.append("DURATION=" + number_to_string(self.duration))
        if self.planned_duration:
            daterange.append(
                "PLANNED-DURATION=" + number_to_string(self.planned_duration)
            )
        if self.scte35_cmd:
            daterange.append("SCTE35-CMD=" + self.scte35_cmd)
        if self.scte35_out:
            daterange.append("SCTE35-OUT=" + self.scte35_out)
        if self.scte35_in:
            daterange.append("SCTE35-IN=" + self.scte35_in)
        if self.end_on_next:
            daterange.append("END-ON-NEXT=" + self.end_on_next)

        # client attributes sorted alphabetically output order is predictable
        for attr, value in sorted(self.x_client_attrs):
            daterange.append("%s=%s" % (denormalize_attribute(attr), value))

        return "#EXT-X-DATERANGE:" + ",".join(daterange)

    def __str__(self):
        return self.dumps()


class ContentSteering(BasePathMixin):
    def __init__(self, base_uri, server_uri, pathway_id=None):
        self.base_uri = base_uri
        self.uri = server_uri
        self.pathway_id = pathway_id

    def dumps(self):
        steering = []
        steering.append("SERVER-URI=" + quoted(self.uri))

        if self.pathway_id is not None:
            steering.append("PATHWAY-ID=" + quoted(self.pathway_id))

        return "#EXT-X-CONTENT-STEERING:" + ",".join(steering)

    def __str__(self):
        return self.dumps()


class ImagePlaylist(BasePathMixin):
    """
    ImagePlaylist object representing a link to a
    variant M3U8 image playlist with a specific bitrate.

    Attributes:

    `image_stream_info` is a named tuple containing the attributes:
     `bandwidth`, `resolution` which is a tuple (w, h) of integers and `codecs`,

    More info: https://github.com/image-media-playlist/spec/blob/master/image_media_playlist_v0_4.pdf
    """

    def __init__(self, base_uri, uri, image_stream_info):
        self.uri = uri
        self.base_uri = base_uri

        resolution = image_stream_info.get("resolution")
        if resolution is not None:
            values = resolution.split("x")
            resolution_pair = (int(values[0]), int(values[1]))
        else:
            resolution_pair = None

        self.image_stream_info = StreamInfo(
            bandwidth=image_stream_info.get("bandwidth"),
            average_bandwidth=image_stream_info.get("average_bandwidth"),
            video=image_stream_info.get("video"),
            # Audio, subtitles, closed captions, video range and hdcp level should not exist in
            # EXT-X-IMAGE-STREAM-INF, so just hardcode them to None.
            audio=None,
            subtitles=None,
            closed_captions=None,
            program_id=image_stream_info.get("program_id"),
            resolution=resolution_pair,
            codecs=image_stream_info.get("codecs"),
            video_range=None,
            hdcp_level=None,
            frame_rate=None,
            pathway_id=image_stream_info.get("pathway_id"),
            stable_variant_id=image_stream_info.get("stable_variant_id")
        )

    def __str__(self):
        image_stream_inf = []
        if self.image_stream_info.program_id:
            image_stream_inf.append("PROGRAM-ID=%d" %
                                     self.image_stream_info.program_id)
        if self.image_stream_info.bandwidth:
            image_stream_inf.append("BANDWIDTH=%d" %
                                     self.image_stream_info.bandwidth)
        if self.image_stream_info.average_bandwidth:
            image_stream_inf.append("AVERAGE-BANDWIDTH=%d" %
                                     self.image_stream_info.average_bandwidth)
        if self.image_stream_info.resolution:
            res = (str(self.image_stream_info.resolution[0]) + "x" +
                   str(self.image_stream_info.resolution[1]))
            image_stream_inf.append("RESOLUTION=" + res)
        if self.image_stream_info.codecs:
            image_stream_inf.append("CODECS=" +
                                     quoted(self.image_stream_info.codecs))
        if self.uri:
            image_stream_inf.append("URI=" + quoted(self.uri))
        if self.image_stream_info.pathway_id:
            image_stream_inf.append(
                "PATHWAY-ID=" + quoted(self.image_stream_info.pathway_id)
            )
        if self.image_stream_info.stable_variant_id:
            image_stream_inf.append(
                "STABLE-VARIANT-ID=" + quoted(self.image_stream_info.stable_variant_id)
            )

        return "#EXT-X-IMAGE-STREAM-INF:" + ",".join(image_stream_inf)

class Tiles(BasePathMixin):
    """
    Image tiles from a M3U8 playlist

    `resolution`
      resolution attribute from EXT-X-TILES tag

    `layout`
      layout attribute from EXT-X-TILES tag

    `duration`
      duration attribute from EXT-X-TILES tag
    """

    def __init__(self, resolution, layout, duration):
        self.resolution = resolution
        self.layout = layout
        self.duration = duration

    def dumps(self):
        tiles = []
        tiles.append("RESOLUTION=" + self.resolution)
        tiles.append("LAYOUT=" + self.layout)
        tiles.append("DURATION=" + self.duration)

        return "#EXT-X-TILES:" + ",".join(tiles)

    def __str__(self):
        return self.dumps()

class ImagePlaylist(BasePathMixin):
    '''
    ImagePlaylist object representing a link to a
    variant M3U8 image playlist with a specific bitrate.

    Attributes:

    `image_stream_info` is a named tuple containing the attributes:
     `bandwidth`, `resolution` which is a tuple (w, h) of integers and `codecs`,

    More info: https://github.com/image-media-playlist/spec/blob/master/image_media_playlist_v0_4.pdf
    '''

    def __init__(self, base_uri, uri, image_stream_info):
        self.uri = uri
        self.base_uri = base_uri

        resolution = image_stream_info.get('resolution')
        if resolution is not None:
            values = resolution.split('x')
            resolution_pair = (int(values[0]), int(values[1]))
        else:
            resolution_pair = None

        self.image_stream_info = StreamInfo(
            bandwidth=image_stream_info.get('bandwidth'),
            average_bandwidth=image_stream_info.get('average_bandwidth'),
            video=image_stream_info.get('video'),
            # Audio, subtitles, closed captions, video range and hdcp level should not exist in
            # EXT-X-IMAGE-STREAM-INF, so just hardcode them to None.
            audio=None,
            subtitles=None,
            closed_captions=None,
            program_id=image_stream_info.get('program_id'),
            resolution=resolution_pair,
            codecs=image_stream_info.get('codecs'),
            video_range=None,
            hdcp_level=None,
            frame_rate=None,
            pathway_id=image_stream_info.get('pathway_id'),
            stable_variant_id=image_stream_info.get('stable_variant_id')
        )

    def __str__(self):
        image_stream_inf = []
        if self.image_stream_info.program_id:
            image_stream_inf.append('PROGRAM-ID=%d' %
                                     self.image_stream_info.program_id)
        if self.image_stream_info.bandwidth:
            image_stream_inf.append('BANDWIDTH=%d' %
                                     self.image_stream_info.bandwidth)
        if self.image_stream_info.average_bandwidth:
            image_stream_inf.append('AVERAGE-BANDWIDTH=%d' %
                                     self.image_stream_info.average_bandwidth)
        if self.image_stream_info.resolution:
            res = (str(self.image_stream_info.resolution[0]) + 'x' +
                   str(self.image_stream_info.resolution[1]))
            image_stream_inf.append('RESOLUTION=' + res)
        if self.image_stream_info.codecs:
            image_stream_inf.append('CODECS=' +
                                     quoted(self.image_stream_info.codecs))
        if self.uri:
            image_stream_inf.append('URI=' + quoted(self.uri))
        if self.image_stream_info.pathway_id:
            image_stream_inf.append(
                'PATHWAY-ID=' + quoted(self.image_stream_info.pathway_id)
            )
        if self.image_stream_info.stable_variant_id:
            image_stream_inf.append(
                'STABLE-VARIANT-ID=' + quoted(self.image_stream_info.stable_variant_id)
            )

        return '#EXT-X-IMAGE-STREAM-INF:' + ','.join(image_stream_inf)


def find_key(keydata, keylist):
    if not keydata:
        return None
    for key in keylist:
        if key:
            # Check the intersection of keys and values
            if (
                keydata.get("uri", None) == key.uri
                and keydata.get("method", "NONE") == key.method
                and keydata.get("iv", None) == key.iv
            ):
                return key
    raise KeyError("No key found for key data")


def denormalize_attribute(attribute):
    return attribute.replace("_", "-").upper()


def quoted(string):
    return '"%s"' % string


def number_to_string(number):
    with decimal.localcontext() as ctx:
        ctx.prec = 20  # set floating point precision
        d = decimal.Decimal(str(number))
        return str(
            d.quantize(decimal.Decimal(1))
            if d == d.to_integral_value()
            else d.normalize()
        )



# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.

import os
import sys
from urllib.parse import urljoin, urlsplit

from m3u8.httpclient import DefaultHTTPClient
from m3u8.model import (
    M3U8,
    ContentSteering,
    DateRange,
    DateRangeList,
    IFramePlaylist,
    ImagePlaylist,
    Key,
    Media,
    MediaList,
    PartialSegment,
    PartialSegmentList,
    PartInformation,
    Playlist,
    PlaylistList,
    PreloadHint,
    RenditionReport,
    RenditionReportList,
    Segment,
    SegmentList,
    ServerControl,
    Skip,
    Start,
    Tiles,
)
from m3u8.parser import ParseError, parse

__all__ = (
    "M3U8",
    "Segment",
    "SegmentList",
    "PartialSegment",
    "PartialSegmentList",
    "Key",
    "Playlist",
    "IFramePlaylist",
    "Media",
    "MediaList",
    "PlaylistList",
    "Start",
    "RenditionReport",
    "RenditionReportList",
    "ServerControl",
    "Skip",
    "PartInformation",
    "PreloadHint",
    "DateRange",
    "DateRangeList",
    "ContentSteering",
    "ImagePlaylist",
    "Tiles",
    "loads",
    "load",
    "parse",
    "ParseError"
)


def loads(content, uri=None, custom_tags_parser=None):
    """
    Given a string with a m3u8 content, returns a M3U8 object.
    Optionally parses a uri to set a correct base_uri on the M3U8 object.
    Raises ValueError if invalid content
    """

    if uri is None:
        return M3U8(content, custom_tags_parser=custom_tags_parser)
    else:
        base_uri = urljoin(uri, ".")
        return M3U8(content, base_uri=base_uri, custom_tags_parser=custom_tags_parser)


def load(
    uri,
    timeout=None,
    headers={},
    custom_tags_parser=None,
    http_client=DefaultHTTPClient(),
    verify_ssl=True,
):
    """
    Retrieves the content from a given URI and returns a M3U8 object.
    Raises ValueError if invalid content or IOError if request fails.
    """
    if urlsplit(uri).scheme:
        content, base_uri = http_client.download(uri, timeout, headers, verify_ssl)
        return M3U8(content, base_uri=base_uri, custom_tags_parser=custom_tags_parser)
    else:
        return _load_from_file(uri, custom_tags_parser)


def _load_from_file(uri, custom_tags_parser=None):
    with open(uri, encoding="utf8") as fileobj:
        raw_content = fileobj.read().strip()
    base_uri = os.path.dirname(uri)
    return M3U8(raw_content, base_uri=base_uri, custom_tags_parser=custom_tags_parser)



from os.path import dirname
from urllib.parse import urljoin, urlsplit


class BasePathMixin(object):
    @property
    def absolute_uri(self):
        if self.uri is None:
            return None

        ret = urljoin(self.base_uri, self.uri)
        if self.base_uri and (not urlsplit(self.base_uri).scheme):
            return ret

        if not urlsplit(ret).scheme:
            raise ValueError("There can not be `absolute_uri` with no `base_uri` set")

        return ret

    @property
    def base_path(self):
        if self.uri is None:
            return None
        return dirname(self.get_path_from_uri())

    def get_path_from_uri(self):
        """Some URIs have a slash in the query string."""
        return self.uri.split("?")[0]

    @base_path.setter
    def base_path(self, newbase_path):
        if self.uri is not None:
            if not self.base_path:
                self.uri = "%s/%s" % (newbase_path, self.uri)
            else:
                self.uri = self.uri.replace(self.base_path, newbase_path)


class GroupedBasePathMixin(object):
    def _set_base_uri(self, new_base_uri):
        for item in self:
            item.base_uri = new_base_uri

    base_uri = property(None, _set_base_uri)

    def _set_base_path(self, newbase_path):
        for item in self:
            item.base_path = newbase_path

    base_path = property(None, _set_base_path)



# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.

import datetime
import itertools
import re
from urllib.parse import urljoin as _urljoin

import iso8601

from m3u8 import protocol

"""
http://tools.ietf.org/html/draft-pantos-http-live-streaming-08#section-3.2
http://stackoverflow.com/questions/2785755/how-to-split-but-ignore-separators-in-quoted-strings-in-python
"""
ATTRIBUTELISTPATTERN = re.compile(r"""((?:[^,"']|"[^"]*"|'[^']*')+)""")


def cast_date_time(value):
    return iso8601.parse_date(value)


def format_date_time(value, **kwargs):
    return value.isoformat(**kwargs)


class ParseError(Exception):
    def __init__(self, lineno, line):
        self.lineno = lineno
        self.line = line

    def __str__(self):
        return "Syntax error in manifest on line %d: %s" % (self.lineno, self.line)


def parse(content, strict=False, custom_tags_parser=None):
    """
    Given a M3U8 playlist content returns a dictionary with all data found
    """
    data = {
        "media_sequence": 0,
        "is_variant": False,
        "is_endlist": False,
        "is_i_frames_only": False,
        "is_independent_segments": False,
        "is_images_only": False,
        "playlist_type": None,
        "playlists": [],
        "segments": [],
        "iframe_playlists": [],
        "image_playlists": [],
        "tiles": [],
        "media": [],
        "keys": [],
        "rendition_reports": [],
        "skip": {},
        "part_inf": {},
        "session_data": [],
        "session_keys": [],
        "segment_map": [],
    }

    state = {
        "expect_segment": False,
        "expect_playlist": False,
        "current_key": None,
        "current_segment_map": None,
    }

    lineno = 0
    for line in string_to_lines(content):
        lineno += 1
        line = line.strip()

        # Call custom parser if needed
        if line.startswith("#") and callable(custom_tags_parser):
            go_to_next_line = custom_tags_parser(line, lineno, data, state)

            # Do not try to parse other standard tags on this line if custom_tags_parser function returns 'True'
            if go_to_next_line:
                continue

        if line.startswith(protocol.ext_x_byterange):
            _parse_byterange(line, state)
            state["expect_segment"] = True
            continue

        if line.startswith(protocol.ext_x_bitrate):
            _parse_bitrate(line, state)

        elif line.startswith(protocol.ext_x_targetduration):
            _parse_simple_parameter(line, data, float)

        elif line.startswith(protocol.ext_x_media_sequence):
            _parse_simple_parameter(line, data, int)

        elif line.startswith(protocol.ext_x_discontinuity_sequence):
            _parse_simple_parameter(line, data, int)

        elif line.startswith(protocol.ext_x_program_date_time):
            _, program_date_time = _parse_simple_parameter_raw_value(
                line, cast_date_time
            )
            if not data.get("program_date_time"):
                data["program_date_time"] = program_date_time
            state["current_program_date_time"] = program_date_time
            state["program_date_time"] = program_date_time

        elif line.startswith(protocol.ext_x_discontinuity):
            state["discontinuity"] = True

        elif line.startswith(protocol.ext_x_cue_out_cont):
            _parse_cueout_cont(line, state)
            state["cue_out"] = True

        elif line.startswith(protocol.ext_x_cue_out):
            _parse_cueout(line, state)
            state["cue_out_start"] = True
            state["cue_out"] = True

        elif line.startswith(f"{protocol.ext_oatcls_scte35}:"):
            _parse_oatcls_scte35(line, state)

        elif line.startswith(f"{protocol.ext_x_asset}:"):
            _parse_asset(line, state)

        elif line.startswith(protocol.ext_x_cue_in):
            state["cue_in"] = True

        elif line.startswith(protocol.ext_x_cue_span):
            state["cue_out"] = True

        elif line.startswith(protocol.ext_x_version):
            _parse_simple_parameter(line, data, int)

        elif line.startswith(protocol.ext_x_allow_cache):
            _parse_simple_parameter(line, data)

        elif line.startswith(protocol.ext_x_key):
            key = _parse_key(line)
            state["current_key"] = key
            if key not in data["keys"]:
                data["keys"].append(key)

        elif line.startswith(protocol.extinf):
            _parse_extinf(line, data, state, lineno, strict)
            state["expect_segment"] = True

        elif line.startswith(protocol.ext_x_stream_inf):
            state["expect_playlist"] = True
            _parse_stream_inf(line, data, state)

        elif line.startswith(protocol.ext_x_i_frame_stream_inf):
            _parse_i_frame_stream_inf(line, data)

        elif line.startswith(protocol.ext_x_media):
            _parse_media(line, data, state)

        elif line.startswith(protocol.ext_x_playlist_type):
            _parse_simple_parameter(line, data)

        elif line.startswith(protocol.ext_i_frames_only):
            data["is_i_frames_only"] = True

        elif line.startswith(protocol.ext_is_independent_segments):
            data["is_independent_segments"] = True

        elif line.startswith(protocol.ext_x_endlist):
            data["is_endlist"] = True

        elif line.startswith(protocol.ext_x_map):
            quoted_parser = remove_quotes_parser("uri", "byterange")
            segment_map_info = _parse_attribute_list(
                protocol.ext_x_map, line, quoted_parser
            )
            state["current_segment_map"] = segment_map_info
            data["segment_map"].append(segment_map_info)

        elif line.startswith(protocol.ext_x_start):
            attribute_parser = {"time_offset": lambda x: float(x)}
            start_info = _parse_attribute_list(
                protocol.ext_x_start, line, attribute_parser
            )
            data["start"] = start_info

        elif line.startswith(protocol.ext_x_server_control):
            _parse_server_control(line, data, state)

        elif line.startswith(protocol.ext_x_part_inf):
            _parse_part_inf(line, data, state)

        elif line.startswith(protocol.ext_x_rendition_report):
            _parse_rendition_report(line, data, state)

        elif line.startswith(protocol.ext_x_part):
            _parse_part(line, data, state)

        elif line.startswith(protocol.ext_x_skip):
            _parse_skip(line, data, state)

        elif line.startswith(protocol.ext_x_session_data):
            _parse_session_data(line, data, state)

        elif line.startswith(protocol.ext_x_session_key):
            _parse_session_key(line, data, state)

        elif line.startswith(protocol.ext_x_preload_hint):
            _parse_preload_hint(line, data, state)

        elif line.startswith(protocol.ext_x_daterange):
            _parse_daterange(line, data, state)

        elif line.startswith(protocol.ext_x_gap):
            state["gap"] = True

        elif line.startswith(protocol.ext_x_content_steering):
            _parse_content_steering(line, data, state)

        elif line.startswith(protocol.ext_x_image_stream_inf):
            _parse_image_stream_inf(line, data)

        elif line.startswith(protocol.ext_x_images_only):
            data['is_images_only'] = True

        elif line.startswith(protocol.ext_x_tiles):
            _parse_tiles(line, data, state)

        elif line.startswith(protocol.ext_m3u):
            # We don't parse #EXTM3U, it just should to be present
            pass

        elif line.strip() == "":
            # blank lines are legal
            pass

        elif (not line.startswith('#')) and (state["expect_segment"]):
            _parse_ts_chunk(line, data, state)
            state["expect_segment"] = False

        elif (not line.startswith('#')) and (state["expect_playlist"]):
            _parse_variant_playlist(line, data, state)
            state["expect_playlist"] = False

        elif strict:
            raise ParseError(lineno, line)

    # there could be remaining partial segments
    if "segment" in state:
        data["segments"].append(state.pop("segment"))

    return data


def _parse_key(line):
    params = ATTRIBUTELISTPATTERN.split(line.replace(protocol.ext_x_key + ":", ""))[
        1::2
    ]
    key = {}
    for param in params:
        name, value = param.split("=", 1)
        key[normalize_attribute(name)] = remove_quotes(value)
    return key


def _parse_extinf(line, data, state, lineno, strict):
    chunks = line.replace(protocol.extinf + ":", "").split(",", 1)
    if len(chunks) == 2:
        duration, title = chunks
    elif len(chunks) == 1:
        if strict:
            raise ParseError(lineno, line)
        else:
            duration = chunks[0]
            title = ""
    if "segment" not in state:
        state["segment"] = {}
    state["segment"]["duration"] = float(duration)
    state["segment"]["title"] = title


def _parse_ts_chunk(line, data, state):
    segment = state.pop("segment")
    if state.get("program_date_time"):
        segment["program_date_time"] = state.pop("program_date_time")
    if state.get("current_program_date_time"):
        segment["current_program_date_time"] = state["current_program_date_time"]
        state["current_program_date_time"] += datetime.timedelta(
            seconds=segment["duration"]
        )
    segment["uri"] = line
    segment["cue_in"] = state.pop("cue_in", False)
    segment["cue_out"] = state.pop("cue_out", False)
    segment["cue_out_start"] = state.pop("cue_out_start", False)
    scte_op = state.pop if segment["cue_in"] else state.get
    segment["scte35"] = scte_op("current_cue_out_scte35", None)
    segment["oatcls_scte35"] = scte_op("current_cue_out_oatcls_scte35", None)
    segment["scte35_duration"] = scte_op("current_cue_out_duration", None)
    segment["scte35_elapsedtime"] = scte_op("current_cue_out_elapsedtime", None)
    segment["asset_metadata"] = scte_op("asset_metadata", None)
    segment["discontinuity"] = state.pop("discontinuity", False)
    if state.get("current_key"):
        segment["key"] = state["current_key"]
    else:
        # For unencrypted segments, the initial key would be None
        if None not in data["keys"]:
            data["keys"].append(None)
    if state.get("current_segment_map"):
        segment["init_section"] = state["current_segment_map"]
    segment["dateranges"] = state.pop("dateranges", None)
    segment["gap_tag"] = state.pop("gap", None)
    data["segments"].append(segment)


def _parse_attribute_list(prefix, line, attribute_parser, default_parser=None):
    params = ATTRIBUTELISTPATTERN.split(line.replace(prefix + ":", ""))[1::2]

    attributes = {}
    for param in params:
        name, value = param.split("=", 1)
        name = normalize_attribute(name)

        if name in attribute_parser:
            value = attribute_parser[name](value)
        elif default_parser is not None:
            value = default_parser(value)

        attributes[name] = value

    return attributes


def _parse_stream_inf(line, data, state):
    data["is_variant"] = True
    data["media_sequence"] = None
    attribute_parser = remove_quotes_parser(
        "codecs",
        "audio",
        "video",
        "video_range",
        "subtitles",
        "pathway_id",
        "stable_variant_id",
    )
    attribute_parser["program_id"] = int
    attribute_parser["bandwidth"] = lambda x: int(float(x))
    attribute_parser["average_bandwidth"] = int
    attribute_parser["frame_rate"] = float
    attribute_parser["video_range"] = str
    attribute_parser["hdcp_level"] = str
    state["stream_info"] = _parse_attribute_list(
        protocol.ext_x_stream_inf, line, attribute_parser
    )


def _parse_i_frame_stream_inf(line, data):
    attribute_parser = remove_quotes_parser(
        "codecs", "uri", "pathway_id", "stable_variant_id"
    )
    attribute_parser["program_id"] = int
    attribute_parser["bandwidth"] = int
    attribute_parser["average_bandwidth"] = int
    attribute_parser["video_range"] = str
    attribute_parser["hdcp_level"] = str
    iframe_stream_info = _parse_attribute_list(
        protocol.ext_x_i_frame_stream_inf, line, attribute_parser
    )
    iframe_playlist = {
        "uri": iframe_stream_info.pop("uri"),
        "iframe_stream_info": iframe_stream_info,
    }

    data["iframe_playlists"].append(iframe_playlist)


def _parse_image_stream_inf(line, data):
    attribute_parser = remove_quotes_parser(
        "codecs", "uri", "pathway_id", "stable_variant_id"
    )
    attribute_parser["program_id"] = int
    attribute_parser["bandwidth"] = int
    attribute_parser["average_bandwidth"] = int
    attribute_parser["resolution"] = str
    image_stream_info = _parse_attribute_list(
        protocol.ext_x_image_stream_inf, line, attribute_parser
    )
    image_playlist = {
        "uri": image_stream_info.pop("uri"),
        "image_stream_info": image_stream_info
    }

    data["image_playlists"].append(image_playlist)



def _parse_tiles(line, data, state):
    attribute_parser = remove_quotes_parser("uri")
    attribute_parser["resolution"] = str
    attribute_parser["layout"] = str
    attribute_parser["duration"] = float
    tiles_info = _parse_attribute_list(
        protocol.ext_x_tiles, line, attribute_parser
    )
    data["tiles"].append(tiles_info)


def _parse_media(line, data, state):
    quoted = remove_quotes_parser(
        "uri",
        "group_id",
        "language",
        "assoc_language",
        "name",
        "instream_id",
        "characteristics",
        "channels",
        "stable_rendition_id",
        "thumbnails",
        "image",
    )
    media = _parse_attribute_list(protocol.ext_x_media, line, quoted)
    data["media"].append(media)


def _parse_variant_playlist(line, data, state):
    playlist = {"uri": line, "stream_info": state.pop("stream_info")}

    data["playlists"].append(playlist)


def _parse_bitrate(line, state):
    if "segment" not in state:
        state["segment"] = {}
    state["segment"]["bitrate"] = line.replace(protocol.ext_x_bitrate + ":", "")


def _parse_byterange(line, state):
    if "segment" not in state:
        state["segment"] = {}
    state["segment"]["byterange"] = line.replace(protocol.ext_x_byterange + ":", "")


def _parse_simple_parameter_raw_value(line, cast_to=str, normalize=False):
    param, value = line.split(":", 1)
    param = normalize_attribute(param.replace("#EXT-X-", ""))
    if normalize:
        value = value.strip().lower()
    return param, cast_to(value)


def _parse_and_set_simple_parameter_raw_value(line, data, cast_to=str, normalize=False):
    param, value = _parse_simple_parameter_raw_value(line, cast_to, normalize)
    data[param] = value
    return data[param]


def _parse_simple_parameter(line, data, cast_to=str):
    return _parse_and_set_simple_parameter_raw_value(line, data, cast_to, True)


def _parse_cueout_cont(line, state):
    elements = line.split(":", 1)
    if len(elements) != 2:
        return

    cue_info = _parse_attribute_list(
        protocol.ext_x_cue_out_cont,
        line,
        remove_quotes_parser("duration", "elapsedtime", "scte35"),
    )

    duration = cue_info.get("duration")
    if duration:
        state["current_cue_out_duration"] = duration

    scte35 = cue_info.get("scte35")
    if duration:
        state["current_cue_out_scte35"] = scte35

    elapsedtime = cue_info.get("elapsedtime")
    if elapsedtime:
        state["current_cue_out_elapsedtime"] = elapsedtime


def _cueout_no_duration(line):
    # this needs to be called first since line.split in all other
    # parsers will throw a ValueError if passed just this tag
    if line == protocol.ext_x_cue_out:
        return (None, None)


def _cueout_envivio(line, state):
    param, value = line.split(":", 1)
    res = re.match('.*DURATION=(.*),.*,CUE="(.*)"', value)
    if res:
        return (res.group(2), res.group(1))
    else:
        return None


def _cueout_duration(line):
    # This was added separately rather than modifying "simple"
    param, value = line.split(":", 1)
    res = re.match(r"DURATION=(.*)", value)
    if res:
        return (None, res.group(1))


def _cueout_simple(line):
    param, value = line.split(":", 1)
    res = re.match(r"^(\d+(?:\.\d)?\d*)$", value)
    if res:
        return (None, res.group(1))


def _parse_cueout(line, state):
    _cueout_state = (
        _cueout_no_duration(line)
        or _cueout_envivio(line, state)
        or _cueout_duration(line)
        or _cueout_simple(line)
    )
    if _cueout_state:
        cue_out_scte35, cue_out_duration = _cueout_state
        current_cue_out_scte35 = state.get("current_cue_out_scte35")
        state["current_cue_out_scte35"] = cue_out_scte35 or current_cue_out_scte35
        state["current_cue_out_duration"] = cue_out_duration


def _parse_server_control(line, data, state):
    attribute_parser = {
        "can_block_reload": str,
        "hold_back": lambda x: float(x),
        "part_hold_back": lambda x: float(x),
        "can_skip_until": lambda x: float(x),
        "can_skip_dateranges": str,
    }

    data["server_control"] = _parse_attribute_list(
        protocol.ext_x_server_control, line, attribute_parser
    )


def _parse_part_inf(line, data, state):
    attribute_parser = {"part_target": lambda x: float(x)}

    data["part_inf"] = _parse_attribute_list(
        protocol.ext_x_part_inf, line, attribute_parser
    )


def _parse_rendition_report(line, data, state):
    attribute_parser = remove_quotes_parser("uri")
    attribute_parser["last_msn"] = int
    attribute_parser["last_part"] = int

    rendition_report = _parse_attribute_list(
        protocol.ext_x_rendition_report, line, attribute_parser
    )

    data["rendition_reports"].append(rendition_report)


def _parse_part(line, data, state):
    attribute_parser = remove_quotes_parser("uri")
    attribute_parser["duration"] = lambda x: float(x)
    attribute_parser["independent"] = str
    attribute_parser["gap"] = str
    attribute_parser["byterange"] = str

    part = _parse_attribute_list(protocol.ext_x_part, line, attribute_parser)

    # this should always be true according to spec
    if state.get("current_program_date_time"):
        part["program_date_time"] = state["current_program_date_time"]
        state["current_program_date_time"] += datetime.timedelta(
            seconds=part["duration"]
        )

    part["dateranges"] = state.pop("dateranges", None)
    part["gap_tag"] = state.pop("gap", None)

    if "segment" not in state:
        state["segment"] = {}
    segment = state["segment"]
    if "parts" not in segment:
        segment["parts"] = []

    segment["parts"].append(part)


def _parse_skip(line, data, state):
    attribute_parser = remove_quotes_parser("recently_removed_dateranges")
    attribute_parser["skipped_segments"] = int

    data["skip"] = _parse_attribute_list(protocol.ext_x_skip, line, attribute_parser)


def _parse_session_data(line, data, state):
    quoted = remove_quotes_parser("data_id", "value", "uri", "language")
    session_data = _parse_attribute_list(protocol.ext_x_session_data, line, quoted)
    data["session_data"].append(session_data)


def _parse_session_key(line, data, state):
    params = ATTRIBUTELISTPATTERN.split(
        line.replace(protocol.ext_x_session_key + ":", "")
    )[1::2]
    key = {}
    for param in params:
        name, value = param.split("=", 1)
        key[normalize_attribute(name)] = remove_quotes(value)
    data["session_keys"].append(key)


def _parse_preload_hint(line, data, state):
    attribute_parser = remove_quotes_parser("uri")
    attribute_parser["type"] = str
    attribute_parser["byterange_start"] = int
    attribute_parser["byterange_length"] = int

    data["preload_hint"] = _parse_attribute_list(
        protocol.ext_x_preload_hint, line, attribute_parser
    )


def _parse_daterange(line, date, state):
    attribute_parser = remove_quotes_parser("id", "class", "start_date", "end_date")
    attribute_parser["duration"] = float
    attribute_parser["planned_duration"] = float
    attribute_parser["end_on_next"] = str
    attribute_parser["scte35_cmd"] = str
    attribute_parser["scte35_out"] = str
    attribute_parser["scte35_in"] = str

    parsed = _parse_attribute_list(protocol.ext_x_daterange, line, attribute_parser)

    if "dateranges" not in state:
        state["dateranges"] = []

    state["dateranges"].append(parsed)


def _parse_content_steering(line, data, state):
    attribute_parser = remove_quotes_parser("server_uri", "pathway_id")

    data["content_steering"] = _parse_attribute_list(
        protocol.ext_x_content_steering, line, attribute_parser
    )


def _parse_oatcls_scte35(line, state):
    scte35_cue = line.split(":", 1)[1]
    state["current_cue_out_oatcls_scte35"] = scte35_cue
    state["current_cue_out_scte35"] = scte35_cue


def _parse_asset(line, state):
    # EXT-X-ASSET attribute values may or may not be quoted, and need to be URL-encoded.
    # They are preserved as-is here to prevent loss of information.
    state["asset_metadata"] = _parse_attribute_list(
        protocol.ext_x_asset, line, {}, default_parser=str
    )


def string_to_lines(string):
    return string.strip().splitlines()


def remove_quotes_parser(*attrs):
    return dict(zip(attrs, itertools.repeat(remove_quotes)))


def remove_quotes(string):
    """
    Remove quotes from string.

    Ex.:
      "foo" -> foo
      'foo' -> foo
      'foo  -> 'foo

    """
    quotes = ('"', "'")
    if string.startswith(quotes) and string.endswith(quotes):
        return string[1:-1]
    return string


def normalize_attribute(attribute):
    return attribute.replace("-", "_").lower().strip()


def get_segment_custom_value(state, key, default=None):
    """
    Helper function for getting custom values for Segment
    Are useful with custom_tags_parser
    """
    if "segment" not in state:
        return default
    if "custom_parser_values" not in state["segment"]:
        return default
    return state["segment"]["custom_parser_values"].get(key, default)


def save_segment_custom_value(state, key, value):
    """
    Helper function for saving custom values for Segment
    Are useful with custom_tags_parser
    """
    if "segment" not in state:
        state["segment"] = {}

    if "custom_parser_values" not in state["segment"]:
        state["segment"]["custom_parser_values"] = {}

    state["segment"]["custom_parser_values"][key] = value



# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.

# Tests M3U8 class to make sure all attributes and methods use the correct
# data returned from parser.parse()

import datetime
import os
import sys

import playlists
import pytest

import m3u8
from m3u8.model import (
    DateRange,
    Key,
    Media,
    MediaList,
    PartialSegment,
    PreloadHint,
    RenditionReport,
    Segment,
    SessionData,
    denormalize_attribute,
    find_key,
)
from m3u8.protocol import ext_x_part, ext_x_preload_hint, ext_x_start


class UTC(datetime.tzinfo):
    """tzinfo class used for backwards compatibility reasons.
    Extracted from the official documentation.
    Ref: https://docs.python.org/2/library/datetime.html#datetime.tzinfo.fromutc
    """

    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return datetime.timedelta(0)


utc = UTC()


def test_base_path_playlist_with_slash_in_query_string():
    playlist = m3u8.M3U8(
        playlists.PLAYLIST_WITH_SLASH_IN_QUERY_STRING,
        base_path="http://testvideo.com/foo",
    )
    assert (
        playlist.segments[0].uri
        == "http://testvideo.com/foo/testvideo-1596635509-4769390994-a0e3087c.ts?hdntl=exp=1596678764~acl=/*~data=hdntl~hmac=12345&"
    )


def test_target_duration_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"targetduration": "1234567"})

    assert "1234567" == obj.target_duration


def test_media_sequence_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"media_sequence": 1234567})

    assert 1234567 == obj.media_sequence


def test_program_date_time_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_PROGRAM_DATE_TIME)

    assert (
        datetime.datetime(2014, 8, 13, 13, 36, 33, tzinfo=utc) == obj.program_date_time
    )


def test_program_date_time_attribute_for_each_segment():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_PROGRAM_DATE_TIME)

    first_program_date_time = datetime.datetime(2014, 8, 13, 13, 36, 33, tzinfo=utc)

    # first segment contains both program_date_time and current_program_date_time
    assert obj.segments[0].program_date_time == first_program_date_time
    assert obj.segments[0].current_program_date_time == first_program_date_time

    # other segments contain only current_program_date_time
    for idx, segment in enumerate(obj.segments[1:]):
        assert segment.program_date_time is None
        assert (
            segment.current_program_date_time
            == first_program_date_time + datetime.timedelta(seconds=(idx + 1) * 3)
        )


def test_program_date_time_attribute_with_discontinuity():
    obj = m3u8.M3U8(playlists.DISCONTINUITY_PLAYLIST_WITH_PROGRAM_DATE_TIME)

    first_program_date_time = datetime.datetime(2014, 8, 13, 13, 36, 33, tzinfo=utc)
    discontinuity_program_date_time = datetime.datetime(
        2014, 8, 13, 13, 36, 55, tzinfo=utc
    )

    segments = obj.segments

    # first segment has EXT-X-PROGRAM-DATE-TIME
    assert segments[0].program_date_time == first_program_date_time
    assert segments[0].current_program_date_time == first_program_date_time

    # second segment does not have EXT-X-PROGRAM-DATE-TIME
    assert segments[1].program_date_time is None
    assert segments[
        1
    ].current_program_date_time == first_program_date_time + datetime.timedelta(
        seconds=3
    )

    # segment with EXT-X-DISCONTINUITY also has EXT-X-PROGRAM-DATE-TIME
    assert segments[5].program_date_time == discontinuity_program_date_time
    assert segments[5].current_program_date_time == discontinuity_program_date_time

    # subsequent segment does not have EXT-X-PROGRAM-DATE-TIME
    assert segments[
        6
    ].current_program_date_time == discontinuity_program_date_time + datetime.timedelta(
        seconds=3
    )
    assert segments[6].program_date_time is None


def test_program_date_time_attribute_without_discontinuity():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_PROGRAM_DATE_TIME_WITHOUT_DISCONTINUITY)

    first_program_date_time = datetime.datetime(2019, 6, 10, 0, 5, tzinfo=utc)

    for idx, segment in enumerate(obj.segments):
        program_date_time = first_program_date_time + datetime.timedelta(
            seconds=idx * 6
        )
        assert segment.program_date_time == program_date_time
        assert segment.current_program_date_time == program_date_time


def test_segment_discontinuity_attribute():
    obj = m3u8.M3U8(playlists.DISCONTINUITY_PLAYLIST_WITH_PROGRAM_DATE_TIME)
    segments = obj.segments

    assert segments[0].discontinuity == False
    assert segments[5].discontinuity == True
    assert segments[6].discontinuity == False


def test_segment_cue_out_attribute():
    obj = m3u8.M3U8(playlists.CUE_OUT_PLAYLIST)
    segments = obj.segments

    assert segments[1].cue_out == True
    assert segments[2].cue_out == True
    assert segments[3].cue_out == False


def test_segment_cue_out_start_attribute():
    obj = m3u8.M3U8(playlists.CUE_OUT_WITH_DURATION_PLAYLIST)

    assert obj.segments[0].cue_out_start == True


def test_segment_cue_in_attribute():
    obj = m3u8.M3U8(playlists.CUE_OUT_WITH_DURATION_PLAYLIST)

    assert obj.segments[2].cue_in == True


def test_segment_cue_out_cont_dumps():
    obj = m3u8.M3U8(playlists.CUE_OUT_PLAYLIST)

    result = obj.dumps()
    expected = "#EXT-X-CUE-OUT-CONT\n"
    assert expected in result


def test_segment_cue_out_cont_attributes_dumps():
    obj = m3u8.M3U8(playlists.CUE_OUT_ELEMENTAL_PLAYLIST)

    result = obj.dumps()
    expected = (
        "#EXT-X-CUE-OUT-CONT:"
        "ElapsedTime=7.960,"
        "Duration=50,SCTE35=/DAlAAAAAAAAAP/wFAUAAAABf+//wpiQkv4ARKogAAEBAQAAQ6sodg==\n"
    )
    assert expected in result


def test_segment_oatcls_scte35_dumps():
    obj = m3u8.M3U8(playlists.CUE_OUT_ELEMENTAL_PLAYLIST)
    result = obj.dumps()

    # Only insert OATCLS-SCTE35 at cue out
    cue_out_line = (
        "#EXT-OATCLS-SCTE35:/DAlAAAAAAAAAP/wFAUAAAABf+//wpiQkv4ARKogAAEBAQAAQ6sodg==\n"
    )
    assert result.count(cue_out_line) == 1


def test_segment_cue_out_start_dumps():
    obj = m3u8.M3U8(playlists.CUE_OUT_WITH_DURATION_PLAYLIST)

    result = obj.dumps()
    expected = "#EXT-X-CUE-OUT:11.52\n"
    assert expected in result


def test_segment_cue_out_start_no_duration_dumps():
    obj = m3u8.M3U8(playlists.CUE_OUT_NO_DURATION_PLAYLIST)

    result = obj.dumps()
    expected = "#EXT-X-CUE-OUT\n"
    assert expected in result


def test_segment_cue_out_in_dumps():
    obj = m3u8.M3U8(playlists.CUE_OUT_NO_DURATION_PLAYLIST)

    result = obj.dumps()
    expected = "#EXT-X-CUE-IN\n"
    assert expected in result


def test_segment_elemental_scte35_attribute():
    obj = m3u8.M3U8(playlists.CUE_OUT_ELEMENTAL_PLAYLIST)
    segments = obj.segments
    assert segments[4].cue_out == True
    assert segments[9].cue_out == False
    assert (
        segments[4].scte35 == "/DAlAAAAAAAAAP/wFAUAAAABf+//wpiQkv4ARKogAAEBAQAAQ6sodg=="
    )


def test_segment_envivio_scte35_attribute():
    obj = m3u8.M3U8(playlists.CUE_OUT_ENVIVIO_PLAYLIST)
    segments = obj.segments
    assert segments[3].cue_out == True
    assert (
        segments[4].scte35 == "/DAlAAAENOOQAP/wFAUBAABrf+//N25XDf4B9p/gAAEBAQAAxKni9A=="
    )
    assert (
        segments[5].scte35 == "/DAlAAAENOOQAP/wFAUBAABrf+//N25XDf4B9p/gAAEBAQAAxKni9A=="
    )
    assert segments[7].cue_out == False


def test_segment_unknown_scte35_attribute():
    obj = m3u8.M3U8(playlists.CUE_OUT_INVALID_PLAYLIST)
    assert obj.segments[0].scte35 is None
    assert obj.segments[0].scte35_duration is None


def test_segment_cue_out_no_duration():
    obj = m3u8.M3U8(playlists.CUE_OUT_NO_DURATION_PLAYLIST)
    assert obj.segments[0].cue_out_start == True
    assert obj.segments[2].cue_in == True


def test_segment_asset_metadata_dumps():
    obj = m3u8.M3U8(playlists.CUE_OUT_ELEMENTAL_PLAYLIST)
    result = obj.dumps()

    # Only insert EXT-X-ASSET at cue out
    asset_metadata_line = (
        '#EXT-X-ASSET:GENRE=CV,CAID=12345678,EPISODE="Episode%20Name%20Date",'
        'SEASON="Season%20Name%20and%20Number",SERIES="Series%2520Name"\n'
    )
    assert result.count(asset_metadata_line) == 1


def test_keys_on_clear_playlist():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)

    assert len(obj.keys) == 1
    assert obj.keys[0] is None


def test_keys_on_simple_encrypted_playlist():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS)

    assert len(obj.keys) == 1
    assert obj.keys[0].uri == "https://priv.example.com/key.php?r=52"


def test_key_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    data = {"keys": [{"method": "AES-128", "uri": "/key", "iv": "foobar"}]}
    mock_parser_data(obj, data)

    assert "Key" == obj.keys[0].__class__.__name__
    assert "AES-128" == obj.keys[0].method
    assert "/key" == obj.keys[0].uri
    assert "foobar" == obj.keys[0].iv


def test_key_attribute_on_none():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {})

    assert len(obj.keys) == 0


def test_key_attribute_without_initialization_vector():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"keys": [{"method": "AES-128", "uri": "/key"}]})

    assert "AES-128" == obj.keys[0].method
    assert "/key" == obj.keys[0].uri
    assert None == obj.keys[0].iv


def test_session_keys_on_clear_playlist():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)

    assert len(obj.session_keys) == 0


def test_session_keys_on_simple_encrypted_playlist():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_SESSION_ENCRYPTED_SEGMENTS)

    assert len(obj.session_keys) == 1
    assert obj.session_keys[0].uri == "https://priv.example.com/key.php?r=52"


def test_session_key_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    data = {"session_keys": [{"method": "AES-128", "uri": "/key", "iv": "foobar"}]}
    mock_parser_data(obj, data)

    assert "SessionKey" == obj.session_keys[0].__class__.__name__
    assert "AES-128" == obj.session_keys[0].method
    assert "/key" == obj.session_keys[0].uri
    assert "foobar" == obj.session_keys[0].iv


def test_session_key_attribute_on_none():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {})

    assert len(obj.session_keys) == 0


def test_session_key_attribute_without_initialization_vector():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"session_keys": [{"method": "AES-128", "uri": "/key"}]})

    assert "AES-128" == obj.session_keys[0].method
    assert "/key" == obj.session_keys[0].uri
    assert None == obj.session_keys[0].iv


def test_segments_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(
        obj,
        {
            "segments": [
                {"uri": "/foo/bar-1.ts", "title": "First Segment", "duration": 1500},
                {"uri": "/foo/bar-2.ts", "title": "Second Segment", "duration": 1600},
            ]
        },
    )

    assert 2 == len(obj.segments)

    assert "/foo/bar-1.ts" == obj.segments[0].uri
    assert "First Segment" == obj.segments[0].title
    assert 1500 == obj.segments[0].duration


def test_segments_attribute_without_title():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"segments": [{"uri": "/foo/bar-1.ts", "duration": 1500}]})

    assert 1 == len(obj.segments)

    assert "/foo/bar-1.ts" == obj.segments[0].uri
    assert 1500 == obj.segments[0].duration
    assert None == obj.segments[0].title


def test_segments_attribute_without_duration():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(
        obj, {"segments": [{"uri": "/foo/bar-1.ts", "title": "Segment title"}]}
    )

    assert 1 == len(obj.segments)

    assert "/foo/bar-1.ts" == obj.segments[0].uri
    assert "Segment title" == obj.segments[0].title
    assert None == obj.segments[0].duration


def test_segments_attribute_with_byterange():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(
        obj,
        {
            "segments": [
                {
                    "uri": "/foo/bar-1.ts",
                    "title": "Segment title",
                    "duration": 1500,
                    "byterange": "76242@0",
                }
            ]
        },
    )

    assert 1 == len(obj.segments)

    assert "/foo/bar-1.ts" == obj.segments[0].uri
    assert "Segment title" == obj.segments[0].title
    assert 1500 == obj.segments[0].duration
    assert "76242@0" == obj.segments[0].byterange


def test_segment_attribute_with_multiple_keys():
    obj = m3u8.M3U8(
        playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV_WITH_MULTIPLE_KEYS
    )

    segments = obj.segments
    assert segments[0].key.uri == "/hls-key/key.bin"
    assert segments[1].key.uri == "/hls-key/key.bin"
    assert segments[4].key.uri == "/hls-key/key2.bin"
    assert segments[5].key.uri == "/hls-key/key2.bin"


def test_segment_title_dumps():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_QUOTED_TITLE)

    result = obj.segments[0].dumps(None).strip()
    expected = '#EXTINF:5220,"A sample title"\nhttp://media.example.com/entire.ts'

    assert result == expected


def test_is_variant_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"is_variant": False})
    assert not obj.is_variant

    mock_parser_data(obj, {"is_variant": True})
    assert obj.is_variant


def test_is_endlist_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"is_endlist": False})
    assert not obj.is_endlist

    obj = m3u8.M3U8(playlists.SLIDING_WINDOW_PLAYLIST)
    mock_parser_data(obj, {"is_endlist": True})
    assert obj.is_endlist


def test_is_i_frames_only_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"is_i_frames_only": False})
    assert not obj.is_i_frames_only

    mock_parser_data(obj, {"is_i_frames_only": True})
    assert obj.is_i_frames_only


def test_playlists_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    data = {
        "playlists": [
            {
                "uri": "/url/1.m3u8",
                "stream_info": {
                    "program_id": 1,
                    "bandwidth": 320000,
                    "closed_captions": None,
                    "video": "high",
                },
            },
            {
                "uri": "/url/2.m3u8",
                "stream_info": {
                    "program_id": 1,
                    "bandwidth": 120000,
                    "closed_captions": None,
                    "codecs": "mp4a.40.5",
                    "video": "low",
                },
            },
        ],
        "media": [
            {"type": "VIDEO", "name": "High", "group_id": "high"},
            {
                "type": "VIDEO",
                "name": "Low",
                "group_id": "low",
                "default": "YES",
                "autoselect": "YES",
            },
        ],
    }
    mock_parser_data(obj, data)

    assert 2 == len(obj.playlists)

    assert "/url/1.m3u8" == obj.playlists[0].uri
    assert 1 == obj.playlists[0].stream_info.program_id
    assert 320000 == obj.playlists[0].stream_info.bandwidth
    assert None == obj.playlists[0].stream_info.closed_captions
    assert None == obj.playlists[0].stream_info.codecs

    assert None == obj.playlists[0].media[0].uri
    assert "high" == obj.playlists[0].media[0].group_id
    assert "VIDEO" == obj.playlists[0].media[0].type
    assert None == obj.playlists[0].media[0].language
    assert "High" == obj.playlists[0].media[0].name
    assert None == obj.playlists[0].media[0].default
    assert None == obj.playlists[0].media[0].autoselect
    assert None == obj.playlists[0].media[0].forced
    assert None == obj.playlists[0].media[0].characteristics

    assert "/url/2.m3u8" == obj.playlists[1].uri
    assert 1 == obj.playlists[1].stream_info.program_id
    assert 120000 == obj.playlists[1].stream_info.bandwidth
    assert None == obj.playlists[1].stream_info.closed_captions
    assert "mp4a.40.5" == obj.playlists[1].stream_info.codecs

    assert None == obj.playlists[1].media[0].uri
    assert "low" == obj.playlists[1].media[0].group_id
    assert "VIDEO" == obj.playlists[1].media[0].type
    assert None == obj.playlists[1].media[0].language
    assert "Low" == obj.playlists[1].media[0].name
    assert "YES" == obj.playlists[1].media[0].default
    assert "YES" == obj.playlists[1].media[0].autoselect
    assert None == obj.playlists[1].media[0].forced
    assert None == obj.playlists[1].media[0].characteristics

    assert [] == obj.iframe_playlists


def test_playlists_attribute_without_program_id():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(
        obj,
        {"playlists": [{"uri": "/url/1.m3u8", "stream_info": {"bandwidth": 320000}}]},
    )

    assert 1 == len(obj.playlists)

    assert "/url/1.m3u8" == obj.playlists[0].uri
    assert 320000 == obj.playlists[0].stream_info.bandwidth
    assert None == obj.playlists[0].stream_info.codecs
    assert None == obj.playlists[0].stream_info.program_id


def test_playlists_attribute_with_resolution():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_RESOLUTION)

    assert 2 == len(obj.playlists)
    assert (512, 288) == obj.playlists[0].stream_info.resolution
    assert None == obj.playlists[1].stream_info.resolution


def test_iframe_playlists_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    data = {
        "iframe_playlists": [
            {
                "uri": "/url/1.m3u8",
                "iframe_stream_info": {
                    "program_id": 1,
                    "bandwidth": 320000,
                    "resolution": "320x180",
                    "codecs": "avc1.4d001f",
                },
            },
            {
                "uri": "/url/2.m3u8",
                "iframe_stream_info": {"bandwidth": "120000", "codecs": "avc1.4d400d"},
            },
        ]
    }
    mock_parser_data(obj, data)

    assert 2 == len(obj.iframe_playlists)

    assert "/url/1.m3u8" == obj.iframe_playlists[0].uri
    assert 1 == obj.iframe_playlists[0].iframe_stream_info.program_id
    assert 320000 == obj.iframe_playlists[0].iframe_stream_info.bandwidth
    assert (320, 180) == obj.iframe_playlists[0].iframe_stream_info.resolution
    assert "avc1.4d001f" == obj.iframe_playlists[0].iframe_stream_info.codecs

    assert "/url/2.m3u8" == obj.iframe_playlists[1].uri
    assert None == obj.iframe_playlists[1].iframe_stream_info.program_id
    assert "120000" == obj.iframe_playlists[1].iframe_stream_info.bandwidth
    assert None == obj.iframe_playlists[1].iframe_stream_info.resolution
    assert "avc1.4d400d" == obj.iframe_playlists[1].iframe_stream_info.codecs


def test_version_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"version": 2})
    assert 2 == obj.version

    mock_parser_data(obj, {})
    assert None == obj.version


def test_version_settable_as_int():
    obj = m3u8.loads(playlists.VERSION_PLAYLIST)
    obj.version = 9

    assert "#EXT-X-VERSION:9" in obj.dumps().strip()


def test_version_settable_as_string():
    obj = m3u8.loads(playlists.VERSION_PLAYLIST)
    obj.version = "9"

    assert "#EXT-X-VERSION:9" in obj.dumps().strip()


def test_allow_cache_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    mock_parser_data(obj, {"allow_cache": "no"})
    assert "no" == obj.allow_cache

    mock_parser_data(obj, {})
    assert None == obj.allow_cache


def test_files_attribute_should_list_all_files_including_segments_and_key():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS)
    files = [
        "https://priv.example.com/key.php?r=52",
        "http://media.example.com/fileSequence52-1.ts",
        "http://media.example.com/fileSequence52-2.ts",
        "http://media.example.com/fileSequence52-3.ts",
    ]
    assert files == obj.files


def test_vod_playlist_type_should_be_imported_as_a_simple_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_VOD_PLAYLIST_TYPE)
    assert obj.playlist_type == "vod"


def test_event_playlist_type_should_be_imported_as_a_simple_attribute():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_EVENT_PLAYLIST_TYPE)
    assert obj.playlist_type == "event"


def test_independent_segments_should_be_true():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_INDEPENDENT_SEGMENTS)
    assert obj.is_independent_segments


def test_independent_segments_should_be_false():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_EVENT_PLAYLIST_TYPE)
    assert not obj.is_independent_segments


def test_no_playlist_type_leaves_attribute_empty():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    assert obj.playlist_type is None


def test_dump_playlists_with_resolution():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_RESOLUTION)

    expected = playlists.SIMPLE_PLAYLIST_WITH_RESOLUTION.strip().splitlines()

    assert expected == obj.dumps().strip().splitlines()


def test_dump_should_build_file_with_same_content(tmpdir):
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV)

    expected = playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV_SORTED.replace(
        ", IV", ",IV"
    ).strip()
    filename = str(tmpdir.join("playlist.m3u8"))

    obj.dump(filename)

    assert_file_content(filename, expected)


def test_dump_should_create_sub_directories(tmpdir):
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV)

    expected = playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV_SORTED.replace(
        ", IV", ",IV"
    ).strip()
    filename = str(tmpdir.join("subdir1", "subdir2", "playlist.m3u8"))

    obj.dump(filename)

    assert_file_content(filename, expected)


def test_dump_should_raise_if_create_sub_directories_fails(tmpdir):
    # The first subdirectory is read-only
    subdir_1 = os.path.join(tmpdir, "subdir1")
    os.mkdir(subdir_1, mode=0o400)

    # The file is to be stored in a second subdirectory that's underneath the first
    subdir_2 = os.path.join(subdir_1, "subdir2")
    file_name = os.path.join(subdir_2, "playlist.m3u8")

    # When we try to write it, we'll be prevented from creating the second subdirectory
    with pytest.raises(OSError):
        m3u8.M3U8(playlists.SIMPLE_PLAYLIST).dump(file_name)


def test_dump_should_work_for_variant_streams():
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST)

    expected = playlists.VARIANT_PLAYLIST.replace(", BANDWIDTH", ",BANDWIDTH").strip()

    assert expected == obj.dumps().strip()


def test_dump_should_work_for_variant_playlists_with_iframe_playlists():
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST_WITH_IFRAME_PLAYLISTS)

    expected = playlists.VARIANT_PLAYLIST_WITH_IFRAME_PLAYLISTS.strip()

    assert expected == obj.dumps().strip()


def test_dump_should_work_for_iframe_playlists():
    obj = m3u8.M3U8(playlists.IFRAME_PLAYLIST)

    expected = playlists.IFRAME_PLAYLIST.strip()

    assert expected == obj.dumps().strip()

    obj = m3u8.M3U8(playlists.IFRAME_PLAYLIST2)

    expected = playlists.IFRAME_PLAYLIST.strip()

    # expected that dump will reverse EXTINF and EXT-X-BYTERANGE,
    # hence IFRAME_PLAYLIST dump from IFRAME_PLAYLIST2 parse.
    assert expected == obj.dumps().strip()

    obj = m3u8.M3U8(playlists.IFRAME_PLAYLIST2)

    expected = playlists.IFRAME_PLAYLIST.strip()

    # expected that dump will reverse EXTINF and EXT-X-BYTERANGE,
    # hence IFRAME_PLAYLIST dump from IFRAME_PLAYLIST2 parse.
    assert expected == obj.dumps().strip()


def test_dump_should_include_program_date_time():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_PROGRAM_DATE_TIME)

    assert (
        "EXT-X-PROGRAM-DATE-TIME:2014-08-13T13:36:33.000+00:00" in obj.dumps().strip()
    )


def test_dump_segment_honors_timespec():
    segment = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_PROGRAM_DATE_TIME).segments[0]
    segment_text = segment.dumps(None, timespec="microseconds").strip()

    assert "EXT-X-PROGRAM-DATE-TIME:2014-08-13T13:36:33.000000+00:00" in segment_text


def test_dump_should_not_ignore_zero_duration():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_ZERO_DURATION)

    assert "EXTINF:0" in obj.dumps().strip()
    assert "EXTINF:5220" in obj.dumps().strip()


def test_dump_should_use_decimal_floating_point_for_very_short_durations():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_VERY_SHORT_DURATION)

    assert "EXTINF:5220" in obj.dumps().strip()
    assert "EXTINF:5218.5" in obj.dumps().strip()
    assert "EXTINF:0.000011" in obj.dumps().strip()


def test_dump_should_include_segment_level_program_date_time():
    obj = m3u8.M3U8(playlists.DISCONTINUITY_PLAYLIST_WITH_PROGRAM_DATE_TIME)
    # Tag being expected is in the segment level, not the global one
    assert (
        "#EXT-X-PROGRAM-DATE-TIME:2014-08-13T13:36:55.000+00:00" in obj.dumps().strip()
    )


def test_dump_should_include_segment_level_program_date_time_without_discontinuity():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_PROGRAM_DATE_TIME_WITHOUT_DISCONTINUITY)

    output = obj.dumps().strip()
    assert "#EXT-X-PROGRAM-DATE-TIME:2019-06-10T00:05:00.000+00:00" in output
    assert "#EXT-X-PROGRAM-DATE-TIME:2019-06-10T00:05:06.000+00:00" in output
    assert "#EXT-X-PROGRAM-DATE-TIME:2019-06-10T00:05:12.000+00:00" in output


def test_dump_should_include_map_attributes():
    obj = m3u8.M3U8(playlists.MAP_URI_PLAYLIST_WITH_BYTERANGE)

    assert 'EXT-X-MAP:URI="main.mp4",BYTERANGE="812@0"' in obj.dumps().strip()


def test_multiple_map_attributes():
    obj = m3u8.M3U8(playlists.MULTIPLE_MAP_URI_PLAYLIST)

    assert obj.segments[0].init_section.uri == "init1.mp4"
    assert obj.segments[1].init_section.uri == "init1.mp4"
    assert obj.segments[2].init_section.uri == "init3.mp4"


def test_dump_should_include_multiple_map_attributes(tmpdir):
    obj = m3u8.M3U8(playlists.MULTIPLE_MAP_URI_PLAYLIST)

    output = obj.dump(str(tmpdir.join("d.m3u8")))
    output = obj.dumps().strip()
    assert output.count('#EXT-X-MAP:URI="init1.mp4"') == 1
    assert output.count('#EXT-X-MAP:URI="init3.mp4"') == 1


def test_dump_should_work_for_playlists_using_byteranges():
    obj = m3u8.M3U8(playlists.PLAYLIST_USING_BYTERANGES)

    expected = playlists.PLAYLIST_USING_BYTERANGES.strip()

    assert expected == obj.dumps().strip()


def test_should_dump_with_endlist_tag():
    obj = m3u8.M3U8(playlists.SLIDING_WINDOW_PLAYLIST)
    obj.is_endlist = True

    assert "#EXT-X-ENDLIST" in obj.dumps().splitlines()


def test_should_dump_without_endlist_tag():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)
    obj.is_endlist = False

    expected = playlists.SIMPLE_PLAYLIST.strip().splitlines()
    expected.remove("#EXT-X-ENDLIST")

    assert expected == obj.dumps().strip().splitlines()


def test_should_dump_multiple_keys():
    obj = m3u8.M3U8(
        playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV_WITH_MULTIPLE_KEYS
    )
    expected = (
        playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV_WITH_MULTIPLE_KEYS_SORTED.strip()
    )

    assert expected == obj.dumps().strip()


def test_should_dump_unencrypted_encrypted_keys_together():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED)
    expected = playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED.strip()

    assert expected == obj.dumps().strip()


def test_should_dump_complex_unencrypted_encrypted_keys():
    obj = m3u8.M3U8(
        playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED_NONE
    )
    expected = (
        playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED_NONE.replace(
            'METHOD=NONE,URI=""', "METHOD=NONE"
        ).strip()
    )

    assert expected == obj.dumps().strip()


def test_should_dump_complex_unencrypted_encrypted_keys_no_uri_attr():
    obj = m3u8.M3U8(
        playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED_NONE_AND_NO_URI_ATTR
    )
    expected = (
        playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED_NONE_AND_NO_URI_ATTR.strip()
    )

    assert expected == obj.dumps().strip()


def test_should_dump_session_data():
    obj = m3u8.M3U8(playlists.SESSION_DATA_PLAYLIST)
    expected = playlists.SESSION_DATA_PLAYLIST.strip()

    assert expected == obj.dumps().strip()


def test_should_dump_multiple_session_data():
    obj = m3u8.M3U8(playlists.MULTIPLE_SESSION_DATA_PLAYLIST)
    expected = playlists.MULTIPLE_SESSION_DATA_PLAYLIST.strip()

    assert expected == obj.dumps().strip()


def test_length_segments_by_key():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED)

    assert len(obj.segments.by_key(obj.keys[0])) == 2
    assert len(obj.segments.by_key(obj.keys[1])) == 4
    assert len(obj.segments.by_key(obj.keys[2])) == 2


def test_list_segments_by_key():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED)

    # unencrypted segments
    segments = obj.segments.by_key(None)
    expected = "../../../../hls/streamNum82400.ts\n../../../../hls/streamNum82401.ts"
    output = [segment.uri for segment in segments]
    assert "\n".join(output).strip() == expected.strip()

    # segments for last key
    segments = obj.segments.by_key(obj.keys[2])
    expected = "../../../../hls/streamNum82404.ts\n../../../../hls/streamNum82405.ts"
    output = [segment.uri for segment in segments]
    assert "\n".join(output).strip() == expected.strip()


def test_replace_segment_key():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED)

    # Replace unencrypted segments with new key
    new_key = Key(
        "AES-128", None, "/hls-key/key0.bin", iv="0Xcafe8f758ca555115584bb5b3c687f52"
    )
    for segment in obj.segments.by_key(None):
        segment.key = new_key

    # Check dump
    expected = (
        playlists.PLAYLIST_WITH_MULTIPLE_KEYS_UNENCRYPTED_AND_ENCRYPTED_UPDATED.strip()
    )

    assert obj.dumps().strip() == expected


def test_keyformat_and_keyformatversion():
    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_KEYFORMAT_AND_KEYFORMATVERSIONS)

    result = obj.dumps().strip()
    expected = 'KEYFORMAT="com.apple.streamingkeydelivery",KEYFORMATVERSIONS="1"'

    assert expected in result


def test_should_dump_program_datetime_and_discontinuity():
    obj = m3u8.M3U8(playlists.DISCONTINUITY_PLAYLIST_WITH_PROGRAM_DATE_TIME)
    expected = playlists.DISCONTINUITY_PLAYLIST_WITH_PROGRAM_DATE_TIME.strip()

    assert expected == obj.dumps().strip()


def test_should_normalize_segments_and_key_urls_if_base_path_passed_to_constructor():
    base_path = "http://videoserver.com/hls/live"

    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV, base_path)

    assert obj.base_path == base_path

    expected = (
        playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV_SORTED.replace(", IV", ",IV")
        .replace("../../../../hls", base_path)
        .replace("/hls-key", base_path)
        .strip()
    )

    assert obj.dumps().strip() == expected


def test_should_normalize_session_key_urls_if_base_path_passed_to_constructor():
    base_path = "http://videoserver.com/hls/live"

    obj = m3u8.M3U8(
        playlists.PLAYLIST_WITH_SESSION_ENCRYPTED_SEGMENTS_AND_IV, base_path
    )

    assert obj.base_path == base_path

    expected = (
        playlists.PLAYLIST_WITH_SESSION_ENCRYPTED_SEGMENTS_AND_IV_SORTED.replace(
            ", IV", ",IV"
        )
        .replace("../../../../hls", base_path)
        .replace("/hls-key", base_path)
        .strip()
    )

    assert obj.dumps().strip() == expected


def test_should_normalize_variant_streams_urls_if_base_path_passed_to_constructor():
    base_path = "http://videoserver.com/hls/live"
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST, base_path)

    expected = (
        playlists.VARIANT_PLAYLIST.replace(", BANDWIDTH", ",BANDWIDTH")
        .replace("http://example.com", base_path)
        .strip()
    )

    assert obj.dumps().strip() == expected


def test_should_normalize_segments_and_key_urls_if_base_path_attribute_updated():
    base_path = "http://videoserver.com/hls/live"

    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV)
    obj.base_path = base_path  # update later

    expected = (
        playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV_SORTED.replace(", IV", ",IV")
        .replace("../../../../hls", base_path)
        .replace("/hls-key", base_path)
        .strip()
    )

    assert obj.dumps() == expected


def test_should_normalize_segments_and_key_urls_if_base_path_attribute_updated():
    base_path = "http://videoserver.com/hls/live"

    obj = m3u8.M3U8(playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV)
    obj.base_path = base_path

    expected = (
        playlists.PLAYLIST_WITH_ENCRYPTED_SEGMENTS_AND_IV_SORTED.replace(", IV", ",IV")
        .replace("../../../../hls", base_path)
        .replace("/hls-key", base_path)
        .strip()
    )

    assert obj.dumps().strip() == expected


def test_playlist_type_dumped_to_appropriate_m3u8_field():
    obj = m3u8.M3U8()
    obj.playlist_type = "vod"
    result = obj.dumps()
    expected = "#EXTM3U\n#EXT-X-PLAYLIST-TYPE:VOD\n"
    assert result == expected


def test_empty_playlist_type_is_gracefully_ignored():
    obj = m3u8.M3U8()
    obj.playlist_type = ""
    result = obj.dumps()
    expected = "#EXTM3U\n"
    assert result == expected


def test_none_playlist_type_is_gracefully_ignored():
    obj = m3u8.M3U8()
    obj.playlist_type = None
    result = obj.dumps()
    expected = "#EXTM3U\n"
    assert result == expected


def test_0_media_sequence_added_to_file():
    obj = m3u8.M3U8()
    obj.media_sequence = 0
    result = obj.dumps()
    expected = "#EXTM3U\n"
    assert result == expected


def test_none_media_sequence_gracefully_ignored():
    obj = m3u8.M3U8()
    obj.media_sequence = None
    result = obj.dumps()
    expected = "#EXTM3U\n"
    assert result == expected


def test_0_discontinuity_sequence_added_to_file():
    obj = m3u8.M3U8()
    obj.discontinuity_sequence = 0
    result = obj.dumps()
    expected = "#EXTM3U\n"
    assert result == expected


def test_none_discontinuity_sequence_gracefully_ignored():
    obj = m3u8.M3U8()
    obj.discontinuity_sequence = None
    result = obj.dumps()
    expected = "#EXTM3U\n"
    assert result == expected


def test_non_zero_discontinuity_sequence_added_to_file():
    obj = m3u8.M3U8()
    obj.discontinuity_sequence = 1
    result = obj.dumps()
    expected = "#EXT-X-DISCONTINUITY-SEQUENCE:1"
    assert expected in result


def test_should_correctly_update_base_path_if_its_blank():
    segment = Segment("entire.ts", "http://1.2/")
    assert not segment.base_path
    segment.base_path = "base_path"
    assert "http://1.2/base_path/entire.ts" == segment.absolute_uri


def test_base_path_should_just_return_uri_if_absolute():
    segment = Segment("http://1.2/entire.ts", "")
    assert "http://1.2/entire.ts" == segment.absolute_uri


def test_m3u8_should_propagate_base_uri_to_segments():
    with open(playlists.RELATIVE_PLAYLIST_FILENAME) as f:
        content = f.read()
    obj = m3u8.M3U8(content, base_uri="/any/path")
    assert "/entire1.ts" == obj.segments[0].uri
    assert "/entire1.ts" == obj.segments[0].absolute_uri
    assert "entire4.ts" == obj.segments[3].uri
    assert "/any/path/entire4.ts" == obj.segments[3].absolute_uri
    obj.base_uri = "/any/where/"
    assert "/entire1.ts" == obj.segments[0].uri
    assert "/entire1.ts" == obj.segments[0].absolute_uri
    assert "entire4.ts" == obj.segments[3].uri
    assert "/any/where/entire4.ts" == obj.segments[3].absolute_uri


def test_m3u8_should_propagate_base_uri_to_key():
    with open(playlists.RELATIVE_PLAYLIST_FILENAME) as f:
        content = f.read()
    obj = m3u8.M3U8(content, base_uri="/any/path")
    assert "../key.bin" == obj.keys[0].uri
    assert "/any/key.bin" == obj.keys[0].absolute_uri
    obj.base_uri = "/any/where/"
    assert "../key.bin" == obj.keys[0].uri
    assert "/any/key.bin" == obj.keys[0].absolute_uri


def test_m3u8_should_propagate_base_uri_to_session_key():
    with open(playlists.RELATIVE_PLAYLIST_FILENAME) as f:
        content = f.read()
    obj = m3u8.M3U8(content, base_uri="/any/path")
    assert "../key.bin" == obj.session_keys[0].uri
    assert "/any/key.bin" == obj.session_keys[0].absolute_uri
    obj.base_uri = "/any/where/"
    assert "../key.bin" == obj.session_keys[0].uri
    assert "/any/key.bin" == obj.session_keys[0].absolute_uri


def test_base_path_with_optional_uri_should_do_nothing():
    media = Media(type="AUDIO", group_id="audio-group", name="English")
    assert media.uri is None
    assert media.base_uri is None
    media.base_path = "base_path"
    assert media.absolute_uri is None
    assert media.base_path is None


def test_medialist_uri_method():
    langs = ["English", "French", "German"]
    ml = MediaList()
    for lang in langs:
        ml.append(
            Media(
                type="AUDIO", group_id="audio-group", name=lang, uri=("/%s.m3u8" % lang)
            )
        )

    assert len(ml.uri) == len(langs)
    assert ml.uri[0] == "/%s.m3u8" % langs[0]
    assert ml.uri[1] == "/%s.m3u8" % langs[1]
    assert ml.uri[2] == "/%s.m3u8" % langs[2]


def test_segment_map_uri_attribute():
    obj = m3u8.M3U8(playlists.MAP_URI_PLAYLIST)
    assert obj.segment_map[0].uri == "fileSequence0.mp4"


def test_segment_map_uri_attribute_with_byterange():
    obj = m3u8.M3U8(playlists.MAP_URI_PLAYLIST_WITH_BYTERANGE)
    assert obj.segment_map[0].uri == "main.mp4"
    assert obj.segment_map[0].byterange == "812@0"
    assert obj.segment_map[1].uri == "main2.mp4"
    assert obj.segment_map[1].byterange == "912@0"


def test_start_with_negative_offset():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_START_NEGATIVE_OFFSET)
    assert obj.start.time_offset == -2.0
    assert obj.start.precise is None
    assert ext_x_start + ":TIME-OFFSET=-2.0\n" in obj.dumps()


def test_start_with_precise():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST_WITH_START_PRECISE)
    assert obj.start.time_offset == 10.5
    assert obj.start.precise == "YES"
    assert ext_x_start + ":TIME-OFFSET=10.5,PRECISE=YES\n" in obj.dumps()


def test_playlist_stream_info_contains_group_id_refs():
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST_WITH_VIDEO_CC_SUBS_AND_AUDIO)
    assert len(obj.playlists) == 2
    for pl in obj.playlists:
        assert pl.stream_info.closed_captions == '"cc"'
        assert pl.stream_info.subtitles == "sub"
        assert pl.stream_info.audio == "aud"
        assert pl.stream_info.video == "vid"


def test_should_dump_frame_rate():
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST_WITH_FRAME_RATE)
    expected = playlists.VARIANT_PLAYLIST_WITH_FRAME_RATE.strip()

    assert expected == obj.dumps().strip()


def test_should_round_frame_rate():
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST_WITH_ROUNDABLE_FRAME_RATE)
    expected = playlists.VARIANT_PLAYLIST_WITH_ROUNDED_FRAME_RATE.strip()

    assert expected == obj.dumps().strip()


@pytest.mark.skipif(sys.version_info >= (3,), reason="unicode not available in v3")
def test_m3u8_unicode_method():
    obj = m3u8.M3U8(playlists.SIMPLE_PLAYLIST)

    result = unicode(obj).strip()
    expected = playlists.SIMPLE_PLAYLIST.strip()
    assert result == expected


def test_add_segment_to_playlist():
    obj = m3u8.M3U8()

    obj.add_segment(Segment("entire.ts", "http://1.2/", duration=1))


def test_segment_str_method():
    segment = Segment("entire.ts", "http://1.2/", duration=1)

    expected = "#EXTINF:1,\nentire.ts"
    result = str(segment).strip()

    assert result == expected


def test_attribute_denormaliser():
    result = denormalize_attribute("test_test")
    expected = "TEST-TEST"

    assert result == expected


def test_find_key_throws_when_no_match():
    threw = False
    try:
        find_key(
            {"method": "AES-128", "iv": 0x12345678, "uri": "http://1.2/"},
            [
                # deliberately empty
            ],
        )
    except KeyError as e:
        threw = True
    finally:
        assert threw


def test_ll_playlist():
    obj = m3u8.M3U8(playlists.LOW_LATENCY_DELTA_UPDATE_PLAYLIST)
    obj.base_path = "http://localhost/test_base_path"
    obj.base_uri = "http://localhost/test_base_uri"

    assert len(obj.rendition_reports) == 2
    assert len(obj.segments[2].parts) == 12
    assert (
        ext_x_part
        + ':DURATION=0.33334,URI="http://localhost/test_base_path/filePart271.0.ts"'
    ) in obj.dumps()
    assert (
        ext_x_preload_hint
        + ':TYPE=PART,URI="http://localhost/test_base_path/filePart273.4.ts"'
    ) in obj.dumps()
    assert obj.preload_hint.base_uri == "http://localhost/test_base_uri"


def test_add_rendition_report_to_playlist():
    obj = m3u8.M3U8()

    obj.add_rendition_report(
        RenditionReport(
            base_uri=None, uri="../1M/waitForMSN.php", last_msn=273, last_part=0
        )
    )

    obj.base_path = "http://localhost/test"

    result = obj.dumps()
    expected = '#EXT-X-RENDITION-REPORT:URI="http://localhost/test/waitForMSN.php",LAST-MSN=273,LAST-PART=0'

    assert expected in result


def test_add_part_to_segment():
    obj = Segment(uri="fileSequence271.ts", duration=4.00008)

    obj.add_part(PartialSegment(None, "filePart271.0.ts", 0.33334))

    result = obj.dumps(None)
    expected = '#EXT-X-PART:DURATION=0.33334,URI="filePart271.0.ts"'

    assert expected in result


def test_partial_segment_gap_and_byterange():
    obj = PartialSegment(
        "", "filePart271.0.ts", 0.33334, byterange="9400@376", gap="YES"
    )

    result = obj.dumps(None)
    expected = (
        '#EXT-X-PART:DURATION=0.33334,URI="filePart271.0.ts",BYTERANGE=9400@376,GAP=YES'
    )

    assert result == expected


def test_session_data_with_value():
    obj = SessionData("com.example.value", "example", language="en")

    result = obj.dumps()
    expected = (
        '#EXT-X-SESSION-DATA:DATA-ID="com.example.value",VALUE="example",LANGUAGE="en"'
    )

    assert result == expected


def test_session_data_with_uri():
    obj = SessionData("com.example.value", uri="example.json", language="en")

    result = obj.dumps()
    expected = '#EXT-X-SESSION-DATA:DATA-ID="com.example.value",URI="example.json",LANGUAGE="en"'

    assert result == expected


def test_session_data_cannot_be_created_with_value_and_uri_at_the_same_time():
    obj = SessionData(
        "com.example.value", value="example", uri="example.json", language="en"
    )

    result = obj.dumps()
    expected = (
        '#EXT-X-SESSION-DATA:DATA-ID="com.example.value",VALUE="example",LANGUAGE="en"'
    )

    assert result == expected


def test_endswith_newline():
    obj = m3u8.loads(playlists.SIMPLE_PLAYLIST)

    manifest = obj.dumps()

    assert manifest.endswith("#EXT-X-ENDLIST\n")


def test_init_section_base_path_update():
    obj = m3u8.M3U8(playlists.MULTIPLE_MAP_URI_PLAYLIST)

    assert obj.segments[0].init_section.uri == "init1.mp4"

    obj.base_path = "http://localhost/base_path"
    obj.base_uri = "http://localhost/base_uri"

    assert obj.segments[0].init_section.uri == "http://localhost/base_path/init1.mp4"
    assert obj.segments[0].init_section.base_uri == "http://localhost/base_uri"


def test_iframe_playlists_base_path_update():
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST_WITH_IFRAME_PLAYLISTS)

    assert obj.iframe_playlists[0].uri == "video-800k-iframes.m3u8"
    assert obj.iframe_playlists[0].base_uri is None

    obj.base_path = "http://localhost/base_path"
    obj.base_uri = "http://localhost/base_uri"

    assert (
        obj.iframe_playlists[0].uri
        == "http://localhost/base_path/video-800k-iframes.m3u8"
    )
    assert obj.iframe_playlists[0].base_uri == "http://localhost/base_uri"


def test_partial_segment_base_path_update():
    obj = m3u8.M3U8(playlists.LOW_LATENCY_DELTA_UPDATE_PLAYLIST)

    obj.base_path = "http://localhost/base_path"
    obj.base_uri = "http://localhost/base_uri"

    assert obj.segments[2].parts[0].uri == "http://localhost/base_path/filePart271.0.ts"
    assert obj.segments[2].parts[0].base_uri == "http://localhost/base_uri"


def test_add_preload_hint():
    obj = PreloadHint("PART", "", "filePart273.4.ts", 0)

    result = obj.dumps()
    expected = '#EXT-X-PRELOAD-HINT:TYPE=PART,URI="filePart273.4.ts",BYTERANGE-START=0'

    assert result == expected


def test_add_daterange():
    obj = DateRange(
        id="testid123",
        start_date="2020-03-09T17:19:00Z",
        planned_duration=60,
        x_test_client_attr='"test-attr"',
    )

    result = obj.dumps()
    expected = '#EXT-X-DATERANGE:ID="testid123",START-DATE="2020-03-09T17:19:00Z",PLANNED-DURATION=60,X-TEST-CLIENT-ATTR="test-attr"'

    assert result == expected


def test_daterange_simple():
    obj = m3u8.M3U8(playlists.DATERANGE_SIMPLE_PLAYLIST)

    # note that x-<client-attribute>s are explicitly alphabetically ordered
    # when dumped for predictability, so line below is different from input
    expected = '#EXT-X-DATERANGE:ID="ad3",START-DATE="2016-06-13T11:15:00Z",DURATION=20,X-AD-ID="1234",X-AD-URL="http://ads.example.com/beacon3"'
    result = obj.dumps()

    assert expected in result


def test_daterange_scte_out_and_in():
    obj = m3u8.M3U8(playlists.DATERANGE_SCTE35_OUT_AND_IN_PLAYLIST)

    result = obj.dumps()

    daterange_out = '#EXT-X-DATERANGE:ID="splice-6FFFFFF0",START-DATE="2014-03-05T11:15:00Z",PLANNED-DURATION=59.993,SCTE35-OUT=0xFC002F0000000000FF000014056FFFFFF000E011622DCAFF000052636200000000000A0008029896F50000008700000000'
    daterange_in = '#EXT-X-DATERANGE:ID="splice-6FFFFFF0",DURATION=59.993,SCTE35-IN=0xFC002A0000000000FF00000F056FFFFFF000401162802E6100000000000A0008029896F50000008700000000'

    assert daterange_out in result
    assert daterange_in in result


def test_daterange_enddate_sctecmd():
    obj = m3u8.M3U8(playlists.DATERANGE_ENDDATE_SCTECMD_PLAYLIST)

    result = obj.dumps()
    expected = '#EXT-X-DATERANGE:ID="test_id",START-DATE="2020-03-11T10:51:00Z",CLASS="test_class",END-DATE="2020-03-11T10:52:00Z",DURATION=60,SCTE35-CMD=0xFCINVALIDSECTION'

    assert expected in result


def test_daterange_in_parts():
    obj = m3u8.M3U8(playlists.DATERANGE_IN_PART_PLAYLIST)

    result = obj.dumps()
    expected = '#EXT-X-DATERANGE:ID="test_id",START-DATE="2020-03-10T07:48:02Z",CLASS="test_class",END-ON-NEXT=YES'

    assert expected in result


def test_add_gap():
    obj = m3u8.Segment(uri="fileSequence271.ts", duration=4, gap_tag=True)

    result = str(obj)
    expected = "#EXTINF:4,\n#EXT-X-GAP\nfileSequence271.ts"

    assert result == expected


def test_gap():
    obj = m3u8.M3U8(playlists.GAP_PLAYLIST)

    result = obj.dumps().strip()
    expected = playlists.GAP_PLAYLIST.strip()

    assert result == expected


def test_gap_in_parts():
    obj = m3u8.M3U8(playlists.GAP_IN_PARTS_PLAYLIST)

    result = obj.dumps().strip()
    expected = playlists.GAP_IN_PARTS_PLAYLIST.strip()

    assert result == expected


def test_skip_dateranges():
    obj = m3u8.M3U8(playlists.DELTA_UPDATE_SKIP_DATERANGES_PLAYLIST)

    expected_skip_tag = (
        '#EXT-X-SKIP:SKIPPED-SEGMENTS=16,RECENTLY-REMOVED-DATERANGES="1"'
    )
    expected_server_control_tag = (
        "#EXT-X-SERVER-CONTROL:CAN-SKIP-UNTIL=36,CAN-SKIP-DATERANGES=YES"
    )

    result = obj.dumps().strip()

    assert expected_skip_tag in result
    assert expected_server_control_tag in result


def test_add_skip():
    obj = m3u8.Skip(skipped_segments=30, recently_removed_dateranges="1\t2")

    expected = '#EXT-X-SKIP:SKIPPED-SEGMENTS=30,RECENTLY-REMOVED-DATERANGES="1\t2"'
    result = obj.dumps().strip()

    assert result == expected


def test_content_steering():
    obj = m3u8.M3U8(playlists.CONTENT_STEERING_PLAYLIST)

    expected_content_steering_tag = (
        '#EXT-X-CONTENT-STEERING:SERVER-URI="/steering?video=00012",PATHWAY-ID="CDN-A"'
    )
    result = obj.dumps().strip()

    assert expected_content_steering_tag in result


def test_add_content_steering():
    obj = m3u8.ContentSteering("", "/steering?video=00012", "CDN-A")

    expected = (
        '#EXT-X-CONTENT-STEERING:SERVER-URI="/steering?video=00012",PATHWAY-ID="CDN-A"'
    )
    result = obj.dumps().strip()

    assert result == expected


def test_content_steering_base_path_update():
    obj = m3u8.M3U8(playlists.CONTENT_STEERING_PLAYLIST)
    obj.base_path = "https://another.example.com/"

    assert (
        '#EXT-X-CONTENT-STEERING:SERVER-URI="https://another.example.com/steering?video=00012",PATHWAY-ID="CDN-A"'
        in obj.dumps().strip()
    )


def test_add_content_steering_base_uri_update():
    obj = m3u8.M3U8(playlists.CONTENT_STEERING_PLAYLIST)
    obj.base_uri = "https://yet-another.example.com/"

    assert (
        obj.content_steering.absolute_uri
        == "https://yet-another.example.com/steering?video=00012"
    )


def test_dump_should_work_for_variant_playlists_with_image_playlists():
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST_WITH_IMAGE_PLAYLISTS)

    expected = playlists.VARIANT_PLAYLIST_WITH_IMAGE_PLAYLISTS.strip()

    assert expected == obj.dumps().strip()

def test_dump_should_work_for_variant_playlists_with_image_playlists():
    obj = m3u8.M3U8(playlists.VARIANT_PLAYLIST_WITH_IMAGE_PLAYLISTS)

    expected = playlists.VARIANT_PLAYLIST_WITH_IMAGE_PLAYLISTS.strip()

    assert expected == obj.dumps().strip()

def test_segment_media_sequence():
    obj = m3u8.M3U8(playlists.SLIDING_WINDOW_PLAYLIST)
    assert [s.media_sequence for s in obj.segments] == [2680, 2681, 2682]


# custom asserts


def assert_file_content(filename, expected):
    with open(filename) as fileobj:
        content = fileobj.read().strip()

    assert content == expected


# helpers


def mock_parser_data(m3u8_obj, data):
    data.setdefault("segments", [])
    m3u8_obj.data = data
    m3u8_obj._initialize_attributes()



# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.

ext_m3u = "#EXTM3U"
ext_x_targetduration = "#EXT-X-TARGETDURATION"
ext_x_media_sequence = "#EXT-X-MEDIA-SEQUENCE"
ext_x_discontinuity_sequence = "#EXT-X-DISCONTINUITY-SEQUENCE"
ext_x_program_date_time = "#EXT-X-PROGRAM-DATE-TIME"
ext_x_media = "#EXT-X-MEDIA"
ext_x_playlist_type = "#EXT-X-PLAYLIST-TYPE"
ext_x_key = "#EXT-X-KEY"
ext_x_stream_inf = "#EXT-X-STREAM-INF"
ext_x_version = "#EXT-X-VERSION"
ext_x_allow_cache = "#EXT-X-ALLOW-CACHE"
ext_x_endlist = "#EXT-X-ENDLIST"
extinf = "#EXTINF"
ext_i_frames_only = "#EXT-X-I-FRAMES-ONLY"
ext_x_asset = "#EXT-X-ASSET"
ext_x_bitrate = "#EXT-X-BITRATE"
ext_x_byterange = "#EXT-X-BYTERANGE"
ext_x_i_frame_stream_inf = "#EXT-X-I-FRAME-STREAM-INF"
ext_x_discontinuity = "#EXT-X-DISCONTINUITY"
ext_x_cue_out = "#EXT-X-CUE-OUT"
ext_x_cue_out_cont = "#EXT-X-CUE-OUT-CONT"
ext_x_cue_in = "#EXT-X-CUE-IN"
ext_x_cue_span = "#EXT-X-CUE-SPAN"
ext_oatcls_scte35 = "#EXT-OATCLS-SCTE35"
ext_is_independent_segments = "#EXT-X-INDEPENDENT-SEGMENTS"
ext_x_map = "#EXT-X-MAP"
ext_x_start = "#EXT-X-START"
ext_x_server_control = "#EXT-X-SERVER-CONTROL"
ext_x_part_inf = "#EXT-X-PART-INF"
ext_x_part = "#EXT-X-PART"
ext_x_rendition_report = "#EXT-X-RENDITION-REPORT"
ext_x_skip = "#EXT-X-SKIP"
ext_x_session_data = "#EXT-X-SESSION-DATA"
ext_x_session_key = "#EXT-X-SESSION-KEY"
ext_x_preload_hint = "#EXT-X-PRELOAD-HINT"
ext_x_daterange = "#EXT-X-DATERANGE"
ext_x_gap = "#EXT-X-GAP"
ext_x_content_steering = "#EXT-X-CONTENT-STEERING"
ext_x_image_stream_inf = '#EXT-X-IMAGE-STREAM-INF'
ext_x_images_only = '#EXT-X-IMAGES-ONLY'
ext_x_tiles = '#EXT-X-TILES'
